"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SpokePoolClient = void 0;
var tslib_1 = require("tslib");
var lodash_1 = require("lodash");
var utils_1 = require("../utils");
var FlowUtils_1 = require("../utils/FlowUtils");
var EventUtils_1 = require("../utils/EventUtils");
var ethers_1 = require("ethers");
var constants_1 = require("../constants");
var NetworkUtils_1 = require("../utils/NetworkUtils");
var BaseAbstractClient_1 = require("./BaseAbstractClient");
var SpokePoolClient = (function (_super) {
    tslib_1.__extends(SpokePoolClient, _super);
    function SpokePoolClient(logger, spokePool, hubPoolClient, chainId, deploymentBlock, eventSearchConfig) {
        if (eventSearchConfig === void 0) { eventSearchConfig = { fromBlock: 0, maxBlockLookBack: 0 }; }
        var _this = _super.call(this) || this;
        _this.logger = logger;
        _this.spokePool = spokePool;
        _this.hubPoolClient = hubPoolClient;
        _this.chainId = chainId;
        _this.deploymentBlock = deploymentBlock;
        _this.eventSearchConfig = eventSearchConfig;
        _this.currentTime = 0;
        _this.depositHashes = {};
        _this.depositHashesToFills = {};
        _this.speedUps = {};
        _this.depositRoutes = {};
        _this.tokensBridged = [];
        _this.rootBundleRelays = [];
        _this.relayerRefundExecutions = [];
        _this.earlyDeposits = [];
        _this.queryableEventNames = [];
        _this.earliestDepositIdQueried = Number.MAX_SAFE_INTEGER;
        _this.latestDepositIdQueried = 0;
        _this.firstDepositIdForSpokePool = Number.MAX_SAFE_INTEGER;
        _this.lastDepositIdForSpokePool = Number.MAX_SAFE_INTEGER;
        _this.latestBlockNumber = 0;
        _this.fills = {};
        _this.refundRequests = [];
        _this.firstBlockToSearch = eventSearchConfig.fromBlock;
        _this.latestBlockSearched = eventSearchConfig.fromBlock;
        _this.queryableEventNames = Object.keys(_this._queryableEventNames());
        return _this;
    }
    SpokePoolClient.prototype._queryableEventNames = function () {
        return {
            FundsDeposited: this.spokePool.filters.FundsDeposited(),
            RequestedSpeedUpDeposit: this.spokePool.filters.RequestedSpeedUpDeposit(),
            FilledRelay: this.spokePool.filters.FilledRelay(),
            RefundRequested: this.spokePool.filters.RefundRequested(),
            EnabledDepositRoute: this.spokePool.filters.EnabledDepositRoute(),
            TokensBridged: this.spokePool.filters.TokensBridged(),
            RelayedRootBundle: this.spokePool.filters.RelayedRootBundle(),
            ExecutedRelayerRefundRoot: this.spokePool.filters.ExecutedRelayerRefundRoot(),
        };
    };
    SpokePoolClient.prototype.getDepositsForDestinationChain = function (destinationChainId) {
        return Object.values(this.depositHashes).filter(function (deposit) { return deposit.destinationChainId === destinationChainId; });
    };
    SpokePoolClient.prototype.getDeposits = function () {
        return (0, EventUtils_1.sortEventsAscendingInPlace)(Object.values(this.depositHashes));
    };
    SpokePoolClient.prototype.getTokensBridged = function () {
        return this.tokensBridged;
    };
    SpokePoolClient.prototype.getDepositRoutes = function () {
        return this.depositRoutes;
    };
    SpokePoolClient.prototype.isDepositRouteEnabled = function (originToken, destinationChainId) {
        var _a, _b;
        return (_b = (_a = this.depositRoutes[originToken]) === null || _a === void 0 ? void 0 : _a[destinationChainId]) !== null && _b !== void 0 ? _b : false;
    };
    SpokePoolClient.prototype.getAllOriginTokens = function () {
        return Object.keys(this.depositRoutes);
    };
    SpokePoolClient.prototype.getFills = function () {
        return (0, EventUtils_1.sortEventsAscendingInPlace)(Object.values(this.fills).flat());
    };
    SpokePoolClient.prototype.getFillsForOriginChain = function (originChainId) {
        return this.fills[originChainId] || [];
    };
    SpokePoolClient.prototype.getFillsForRelayer = function (relayer) {
        return this.getFills().filter(function (fill) { return fill.relayer === relayer; });
    };
    SpokePoolClient.prototype.getFillsWithBlockInRange = function (startingBlock, endingBlock) {
        return this.getFills().filter(function (fill) { return fill.blockNumber >= startingBlock && fill.blockNumber <= endingBlock; });
    };
    SpokePoolClient.prototype.getRefundRequests = function (fromBlock, toBlock) {
        return fromBlock === undefined || toBlock === undefined || isNaN(fromBlock) || isNaN(toBlock)
            ? this.refundRequests
            : this.refundRequests.filter(function (request) { return request.blockNumber >= fromBlock && request.blockNumber <= toBlock; });
    };
    SpokePoolClient.prototype.getRootBundleRelays = function () {
        return this.rootBundleRelays;
    };
    SpokePoolClient.prototype.getLatestRootBundleId = function () {
        var _a;
        return this.rootBundleRelays.length > 0
            ? ((_a = this.rootBundleRelays[this.rootBundleRelays.length - 1]) === null || _a === void 0 ? void 0 : _a.rootBundleId) + 1
            : 0;
    };
    SpokePoolClient.prototype.getRelayerRefundExecutions = function () {
        return this.relayerRefundExecutions;
    };
    SpokePoolClient.prototype.getExecutedRefunds = function (relayerRefundRoot) {
        var bundle = this.getRootBundleRelays().find(function (bundle) { return bundle.relayerRefundRoot === relayerRefundRoot; });
        if (bundle === undefined) {
            return {};
        }
        var executedRefundLeaves = this.getRelayerRefundExecutions().filter(function (leaf) { return leaf.rootBundleId === bundle.rootBundleId; });
        var executedRefunds = {};
        for (var _i = 0, executedRefundLeaves_1 = executedRefundLeaves; _i < executedRefundLeaves_1.length; _i++) {
            var refundLeaf = executedRefundLeaves_1[_i];
            var tokenAddress = refundLeaf.l2TokenAddress;
            if (executedRefunds[tokenAddress] === undefined) {
                executedRefunds[tokenAddress] = {};
            }
            var executedTokenRefunds = executedRefunds[tokenAddress];
            for (var i = 0; i < refundLeaf.refundAddresses.length; i++) {
                var relayer = refundLeaf.refundAddresses[i];
                var refundAmount = refundLeaf.refundAmounts[i];
                if (executedTokenRefunds[relayer] === undefined) {
                    executedTokenRefunds[relayer] = ethers_1.ethers.constants.Zero;
                }
                executedTokenRefunds[relayer] = executedTokenRefunds[relayer].add(refundAmount);
            }
        }
        return executedRefunds;
    };
    SpokePoolClient.prototype.appendMaxSpeedUpSignatureToDeposit = function (deposit) {
        var _a, _b;
        var maxSpeedUp = (_b = (_a = this.speedUps[deposit.depositor]) === null || _a === void 0 ? void 0 : _a[deposit.depositId]) === null || _b === void 0 ? void 0 : _b.reduce(function (prev, current) {
            return prev.newRelayerFeePct.gt(current.newRelayerFeePct) ? prev : current;
        });
        if (!maxSpeedUp || maxSpeedUp.newRelayerFeePct.lte(deposit.relayerFeePct)) {
            return deposit;
        }
        return tslib_1.__assign(tslib_1.__assign({}, deposit), { speedUpSignature: maxSpeedUp.depositorSignature, newRelayerFeePct: maxSpeedUp.newRelayerFeePct, updatedRecipient: maxSpeedUp.updatedRecipient, updatedMessage: maxSpeedUp.updatedMessage });
    };
    SpokePoolClient.prototype.getDepositForFill = function (fill, fillFieldsToIgnore) {
        if (fillFieldsToIgnore === void 0) { fillFieldsToIgnore = []; }
        var depositWithMatchingDepositId = this.depositHashes[this.getDepositHash(fill)];
        return (0, FlowUtils_1.validateFillForDeposit)(fill, depositWithMatchingDepositId, fillFieldsToIgnore)
            ? depositWithMatchingDepositId
            : undefined;
    };
    SpokePoolClient.prototype.updateDepositRealizedLpFeePct = function (event, realizedLpFeePct) {
        this.depositHashes[this.getDepositHash(event)] = tslib_1.__assign(tslib_1.__assign({}, this.depositHashes[this.getDepositHash(event)]), { realizedLpFeePct: realizedLpFeePct });
    };
    SpokePoolClient.prototype.getValidUnfilledAmountForDeposit = function (deposit) {
        var fillsForDeposit = this.depositHashesToFills[this.getDepositHash(deposit)];
        if (fillsForDeposit === undefined || fillsForDeposit.length === 0) {
            return { unfilledAmount: (0, utils_1.toBN)(deposit.amount), fillCount: 0, invalidFills: [] };
        }
        var _a = fillsForDeposit.reduce(function (groupedFills, fill) {
            if ((0, FlowUtils_1.validateFillForDeposit)(fill, deposit)) {
                groupedFills.validFills.push(fill);
            }
            else {
                groupedFills.invalidFills.push(fill);
            }
            return groupedFills;
        }, { validFills: [], invalidFills: [] }), validFills = _a.validFills, invalidFills = _a.invalidFills;
        var invalidFillsForDeposit = invalidFills.filter(function (x) { return x.depositId === deposit.depositId; });
        if (invalidFillsForDeposit.length > 0) {
            this.logger.warn({
                at: "SpokePoolClient",
                chainId: this.chainId,
                message: "Invalid fills found matching deposit ID",
                deposit: deposit,
                invalidFills: Object.fromEntries(invalidFillsForDeposit.map(function (x) { return [x.relayer, x]; })),
            });
        }
        if (validFills.length === 0) {
            return { unfilledAmount: (0, utils_1.toBN)(deposit.amount), fillCount: 0, invalidFills: invalidFills };
        }
        var fillsOrderedByTotalFilledAmount = validFills.sort(function (fillA, fillB) {
            return fillB.totalFilledAmount.gt(fillA.totalFilledAmount)
                ? 1
                : fillB.totalFilledAmount.lt(fillA.totalFilledAmount)
                    ? -1
                    : 0;
        });
        var lastFill = fillsOrderedByTotalFilledAmount[0];
        return {
            unfilledAmount: (0, utils_1.toBN)(lastFill.amount.sub(lastFill.totalFilledAmount)),
            fillCount: validFills.length,
            invalidFills: invalidFills,
        };
    };
    SpokePoolClient.prototype.getDepositHash = function (event) {
        return "".concat(event.depositId, "-").concat(event.originChainId);
    };
    SpokePoolClient.prototype._getBlockRangeForDepositId = function (targetDepositId, initLow, initHigh, maxSearches) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var low, high, i, mid, searchedDepositId;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (initLow > initHigh) {
                            throw new Error("Binary search failed because low > high");
                        }
                        if (maxSearches <= 0) {
                            throw new Error("maxSearches must be > 0");
                        }
                        low = initLow;
                        high = initHigh;
                        i = 0;
                        _a.label = 1;
                    case 1:
                        mid = Math.floor((high + low) / 2);
                        return [4, this._getDepositIdAtBlock(mid)];
                    case 2:
                        searchedDepositId = _a.sent();
                        if (!Number.isInteger(searchedDepositId)) {
                            throw new Error("Invalid deposit count");
                        }
                        if (i++ >= maxSearches) {
                            return [2, { low: low, high: high }];
                        }
                        if (targetDepositId > searchedDepositId) {
                            low = mid;
                        }
                        else if (targetDepositId < searchedDepositId) {
                            high = mid;
                        }
                        else {
                            return [2, { low: low, high: high }];
                        }
                        _a.label = 3;
                    case 3:
                        if (low <= high) return [3, 1];
                        _a.label = 4;
                    case 4: throw new Error("Failed to find deposit ID");
                }
            });
        });
    };
    SpokePoolClient.prototype._getDepositIdAtBlock = function (blockTag) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.spokePool.numberOfDeposits({ blockTag: blockTag })];
                    case 1: return [2, _a.sent()];
                }
            });
        });
    };
    SpokePoolClient.prototype.queryHistoricalMatchingFills = function (fill, deposit, toBlock) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var searchConfig;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        searchConfig = {
                            fromBlock: this.deploymentBlock,
                            toBlock: toBlock,
                            maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                        };
                        return [4, this.queryFillsInBlockRange(fill, searchConfig)];
                    case 1: return [2, (_a.sent()).filter(function (_fill) {
                            return (0, FlowUtils_1.validateFillForDeposit)(_fill, deposit);
                        })];
                }
            });
        });
    };
    SpokePoolClient.prototype.queryFillsInBlockRange = function (matchingFill, searchConfig) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var query, fills;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, (0, EventUtils_1.paginatedEventQuery)(this.spokePool, this.spokePool.filters.FilledRelay(undefined, undefined, undefined, undefined, matchingFill.originChainId, undefined, undefined, undefined, matchingFill.depositId, undefined, undefined, matchingFill.depositor, undefined, undefined, undefined), searchConfig)];
                    case 1:
                        query = _a.sent();
                        fills = query.map(function (event) { return (0, EventUtils_1.spreadEventWithBlockNumber)(event); });
                        return [2, (0, EventUtils_1.sortEventsAscending)(fills.filter(function (_fill) { return (0, FlowUtils_1.filledSameDeposit)(_fill, matchingFill); }))];
                }
            });
        });
    };
    SpokePoolClient.prototype._update = function (eventsToQuery) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var firstDepositId, latestBlockNumber, searchConfig, eventSearchConfigs, timerStart, _a, numberOfDeposits, currentTime, events, errMsg, blockNumbers, blocks, _b, _c;
            var _this = this;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        firstDepositId = this.firstDepositIdForSpokePool;
                        if (!(firstDepositId === Number.MAX_SAFE_INTEGER)) return [3, 2];
                        return [4, this.spokePool.numberOfDeposits({ blockTag: this.deploymentBlock })];
                    case 1:
                        firstDepositId = _d.sent();
                        if (isNaN(firstDepositId) || firstDepositId < 0) {
                            throw new Error("SpokePoolClient::update: Invalid first deposit id (".concat(firstDepositId, ")"));
                        }
                        _d.label = 2;
                    case 2: return [4, this.spokePool.provider.getBlockNumber()];
                    case 3:
                        latestBlockNumber = _d.sent();
                        if (isNaN(latestBlockNumber) || latestBlockNumber < this.latestBlockNumber) {
                            throw new Error("SpokePoolClient::update: latestBlockNumber ".concat(latestBlockNumber, " < ").concat(this.latestBlockNumber));
                        }
                        searchConfig = {
                            fromBlock: this.firstBlockToSearch,
                            toBlock: this.eventSearchConfig.toBlock || latestBlockNumber,
                            maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                        };
                        if (searchConfig.fromBlock > searchConfig.toBlock) {
                            this.log("warn", "Invalid update() searchConfig.", { searchConfig: searchConfig });
                            return [2, { success: false }];
                        }
                        eventSearchConfigs = eventsToQuery.map(function (eventName) {
                            if (!_this.queryableEventNames.includes(eventName)) {
                                throw new Error("SpokePoolClient: Cannot query unrecognised SpokePool event name: ".concat(eventName));
                            }
                            var _searchConfig = tslib_1.__assign({}, searchConfig);
                            if (eventName === "EnabledDepositRoute" && !_this.isUpdated) {
                                _searchConfig.fromBlock = _this.deploymentBlock;
                            }
                            return {
                                filter: _this._queryableEventNames()[eventName],
                                searchConfig: _searchConfig,
                            };
                        });
                        this.log("debug", "Updating SpokePool client for chain ".concat(this.chainId), {
                            eventsToQuery: eventsToQuery,
                            searchConfig: searchConfig,
                            spokePool: this.spokePool.address,
                        });
                        timerStart = Date.now();
                        return [4, Promise.all(tslib_1.__spreadArray([
                                this.spokePool.numberOfDeposits({ blockTag: searchConfig.toBlock }),
                                this.spokePool.getCurrentTime({ blockTag: searchConfig.toBlock })
                            ], eventSearchConfigs.map(function (config) { return (0, EventUtils_1.paginatedEventQuery)(_this.spokePool, config.filter, config.searchConfig); }), true))];
                    case 4:
                        _a = _d.sent(), numberOfDeposits = _a[0], currentTime = _a[1], events = _a.slice(2);
                        this.log("debug", "Time to query new events from RPC for ".concat(this.chainId, ": ").concat(Date.now() - timerStart, " ms"));
                        if (!ethers_1.BigNumber.isBigNumber(currentTime) || currentTime.lt(this.currentTime)) {
                            errMsg = ethers_1.BigNumber.isBigNumber(currentTime)
                                ? "currentTime: ".concat(currentTime, " < ").concat((0, utils_1.toBN)(this.currentTime))
                                : "currentTime is not a BigNumber: ".concat(JSON.stringify(currentTime));
                            throw new Error("SpokePoolClient::update: ".concat(errMsg));
                        }
                        events.forEach(function (events) { return (0, EventUtils_1.sortEventsAscendingInPlace)(events); });
                        blockNumbers = Array.from(new Set(["FundsDeposited", "FilledRelay", "RefundRequested"]
                            .filter(function (eventName) { return eventsToQuery.includes(eventName); })
                            .map(function (eventName) {
                            var idx = eventsToQuery.indexOf(eventName);
                            return events[idx].map(function (_a) {
                                var blockNumber = _a.blockNumber;
                                return blockNumber;
                            });
                        })
                            .flat()));
                        _c = (_b = Object).fromEntries;
                        return [4, (0, utils_1.mapAsync)(blockNumbers, function (blockNumber) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                var block;
                                return tslib_1.__generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4, this.spokePool.provider.getBlock(blockNumber)];
                                        case 1:
                                            block = _a.sent();
                                            return [2, [blockNumber, block]];
                                    }
                                });
                            }); })];
                    case 5:
                        blocks = _c.apply(_b, [_d.sent()]);
                        return [2, {
                                success: true,
                                currentTime: currentTime.toNumber(),
                                firstDepositId: firstDepositId,
                                latestBlockNumber: latestBlockNumber,
                                latestDepositId: Math.max(numberOfDeposits - 1, 0),
                                searchEndBlock: searchConfig.toBlock,
                                events: events,
                                blocks: blocks,
                            }];
                }
            });
        });
    };
    SpokePoolClient.prototype._isEarlyDeposit = function (depositEvent, currentTime) {
        return depositEvent.args.quoteTimestamp > currentTime;
    };
    SpokePoolClient.prototype.update = function (eventsToQuery) {
        var _a, _b;
        if (eventsToQuery === void 0) { eventsToQuery = this.queryableEventNames; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var update, queryResults, blocks, currentTime, _i, _c, event_1, allDeposits, _d, _e, earlyDeposits, _f, depositEvents, dataForQuoteTime, _g, _h, _j, index, event_2, rawDeposit, deposit, _k, _l, speedUpEvents, _m, speedUpEvents_1, event_3, speedUp, _o, _p, deposits, fillEvents, _q, fillEvents_1, event_4, rawFill, fill, refundRequests, _r, refundRequests_1, event_5, rawRefundRequest, refundRequest, enableDepositsEvents, _s, enableDepositsEvents_1, event_6, enableDeposit, relayedRootBundleEvents, _t, relayedRootBundleEvents_1, event_7, executedRelayerRefundRootEvents, _u, executedRelayerRefundRootEvents_1, event_8, executedRefund;
            var _v;
            var _this = this;
            return tslib_1.__generator(this, function (_w) {
                switch (_w.label) {
                    case 0:
                        if (this.hubPoolClient !== null && !this.hubPoolClient.isUpdated) {
                            throw new Error("HubPoolClient not updated");
                        }
                        return [4, this._update(eventsToQuery)];
                    case 1:
                        update = _w.sent();
                        if (!update.success) {
                            return [2];
                        }
                        queryResults = update.events, blocks = update.blocks, currentTime = update.currentTime;
                        if (eventsToQuery.includes("TokensBridged")) {
                            for (_i = 0, _c = queryResults[eventsToQuery.indexOf("TokensBridged")]; _i < _c.length; _i++) {
                                event_1 = _c[_i];
                                this.tokensBridged.push((0, EventUtils_1.spreadEventWithBlockNumber)(event_1));
                            }
                        }
                        if (!eventsToQuery.includes("FundsDeposited")) return [3, 8];
                        allDeposits = tslib_1.__spreadArray(tslib_1.__spreadArray([], queryResults[eventsToQuery.indexOf("FundsDeposited")], true), this.earlyDeposits, true);
                        _d = (0, lodash_1.groupBy)(allDeposits, function (depositEvent) {
                            if (_this._isEarlyDeposit(depositEvent, currentTime)) {
                                var args = depositEvent.args, transactionHash = depositEvent.transactionHash;
                                _this.logger.debug({
                                    at: "SpokePoolClient#update",
                                    message: "Deferring early deposit event.",
                                    currentTime: currentTime,
                                    deposit: { args: args, transactionHash: transactionHash },
                                });
                                return "earlyDeposits";
                            }
                            else {
                                return "depositEvents";
                            }
                        }), _e = _d.earlyDeposits, earlyDeposits = _e === void 0 ? [] : _e, _f = _d.depositEvents, depositEvents = _f === void 0 ? [] : _f;
                        this.earlyDeposits = earlyDeposits;
                        return [4, Promise.all(depositEvents.map(function (event) { return tslib_1.__awaiter(_this, void 0, void 0, function () { return tslib_1.__generator(this, function (_a) {
                                return [2, this.computeRealizedLpFeePct(event)];
                            }); }); }))];
                    case 2:
                        dataForQuoteTime = _w.sent();
                        if (depositEvents.length > 0) {
                            this.log("debug", "Using ".concat(depositEvents.length, " newly queried deposit events for chain ").concat(this.chainId), {
                                earliestEvent: depositEvents[0].blockNumber,
                            });
                        }
                        _g = 0, _h = Array.from(depositEvents.entries());
                        _w.label = 3;
                    case 3:
                        if (!(_g < _h.length)) return [3, 8];
                        _j = _h[_g], index = _j[0], event_2 = _j[1];
                        rawDeposit = (0, EventUtils_1.spreadEventWithBlockNumber)(event_2);
                        _k = [tslib_1.__assign({}, rawDeposit)];
                        _v = { realizedLpFeePct: dataForQuoteTime[index].realizedLpFeePct, destinationToken: this.getDestinationTokenForDeposit(rawDeposit), quoteBlockNumber: dataForQuoteTime[index].quoteBlock };
                        if (!((_b = (_a = blocks[event_2.blockNumber]) === null || _a === void 0 ? void 0 : _a.timestamp) !== null && _b !== void 0)) return [3, 4];
                        _l = _b;
                        return [3, 6];
                    case 4: return [4, event_2.getBlock()];
                    case 5:
                        _l = (_w.sent()).timestamp;
                        _w.label = 6;
                    case 6:
                        deposit = tslib_1.__assign.apply(void 0, _k.concat([(_v.blockTimestamp = _l, _v)]));
                        (0, utils_1.assign)(this.depositHashes, [this.getDepositHash(deposit)], deposit);
                        if (deposit.depositId < this.earliestDepositIdQueried) {
                            this.earliestDepositIdQueried = deposit.depositId;
                        }
                        if (deposit.depositId > this.latestDepositIdQueried) {
                            this.latestDepositIdQueried = deposit.depositId;
                        }
                        _w.label = 7;
                    case 7:
                        _g++;
                        return [3, 3];
                    case 8:
                        if (eventsToQuery.includes("RequestedSpeedUpDeposit")) {
                            speedUpEvents = queryResults[eventsToQuery.indexOf("RequestedSpeedUpDeposit")];
                            for (_m = 0, speedUpEvents_1 = speedUpEvents; _m < speedUpEvents_1.length; _m++) {
                                event_3 = speedUpEvents_1[_m];
                                speedUp = tslib_1.__assign(tslib_1.__assign({}, (0, EventUtils_1.spreadEvent)(event_3.args)), { originChainId: this.chainId });
                                (0, utils_1.assign)(this.speedUps, [speedUp.depositor, speedUp.depositId], [speedUp]);
                            }
                            for (_o = 0, _p = Object.entries(this.depositHashes); _o < _p.length; _o++) {
                                deposits = _p[_o];
                                this.depositHashes[deposits[0]] = this.appendMaxSpeedUpSignatureToDeposit(deposits[1]);
                            }
                        }
                        if (eventsToQuery.includes("FilledRelay")) {
                            fillEvents = queryResults[eventsToQuery.indexOf("FilledRelay")];
                            if (fillEvents.length > 0) {
                                this.log("debug", "Using ".concat(fillEvents.length, " newly queried fill events for chain ").concat(this.chainId), {
                                    earliestEvent: fillEvents[0].blockNumber,
                                });
                            }
                            for (_q = 0, fillEvents_1 = fillEvents; _q < fillEvents_1.length; _q++) {
                                event_4 = fillEvents_1[_q];
                                rawFill = (0, EventUtils_1.spreadEventWithBlockNumber)(event_4);
                                fill = tslib_1.__assign(tslib_1.__assign({}, rawFill), { blockTimestamp: blocks[event_4.blockNumber].timestamp });
                                (0, utils_1.assign)(this.fills, [fill.originChainId], [fill]);
                                (0, utils_1.assign)(this.depositHashesToFills, [this.getDepositHash(fill)], [fill]);
                            }
                        }
                        if (eventsToQuery.includes("RefundRequested")) {
                            refundRequests = queryResults[eventsToQuery.indexOf("RefundRequested")];
                            if (refundRequests.length > 0) {
                                this.log("debug", "Found ".concat(refundRequests.length, " new relayer refund requests on chain ").concat(this.chainId), {
                                    earliestEvent: refundRequests[0].blockNumber,
                                });
                            }
                            for (_r = 0, refundRequests_1 = refundRequests; _r < refundRequests_1.length; _r++) {
                                event_5 = refundRequests_1[_r];
                                rawRefundRequest = (0, EventUtils_1.spreadEventWithBlockNumber)(event_5);
                                refundRequest = tslib_1.__assign(tslib_1.__assign({}, rawRefundRequest), { repaymentChainId: this.chainId, blockTimestamp: blocks[event_5.blockNumber].timestamp });
                                this.refundRequests.push(refundRequest);
                            }
                        }
                        if (eventsToQuery.includes("EnabledDepositRoute")) {
                            enableDepositsEvents = queryResults[eventsToQuery.indexOf("EnabledDepositRoute")];
                            for (_s = 0, enableDepositsEvents_1 = enableDepositsEvents; _s < enableDepositsEvents_1.length; _s++) {
                                event_6 = enableDepositsEvents_1[_s];
                                enableDeposit = (0, EventUtils_1.spreadEvent)(event_6.args);
                                (0, utils_1.assign)(this.depositRoutes, [enableDeposit.originToken, enableDeposit.destinationChainId], enableDeposit.enabled);
                            }
                        }
                        if (eventsToQuery.includes("RelayedRootBundle")) {
                            relayedRootBundleEvents = queryResults[eventsToQuery.indexOf("RelayedRootBundle")];
                            for (_t = 0, relayedRootBundleEvents_1 = relayedRootBundleEvents; _t < relayedRootBundleEvents_1.length; _t++) {
                                event_7 = relayedRootBundleEvents_1[_t];
                                this.rootBundleRelays.push((0, EventUtils_1.spreadEventWithBlockNumber)(event_7));
                            }
                        }
                        if (eventsToQuery.includes("ExecutedRelayerRefundRoot")) {
                            executedRelayerRefundRootEvents = queryResults[eventsToQuery.indexOf("ExecutedRelayerRefundRoot")];
                            for (_u = 0, executedRelayerRefundRootEvents_1 = executedRelayerRefundRootEvents; _u < executedRelayerRefundRootEvents_1.length; _u++) {
                                event_8 = executedRelayerRefundRootEvents_1[_u];
                                executedRefund = (0, EventUtils_1.spreadEventWithBlockNumber)(event_8);
                                executedRefund.l2TokenAddress = SpokePoolClient.getExecutedRefundLeafL2Token(executedRefund.chainId, executedRefund.l2TokenAddress);
                                this.relayerRefundExecutions.push(executedRefund);
                            }
                        }
                        this.currentTime = currentTime;
                        this.firstDepositIdForSpokePool = update.firstDepositId;
                        this.latestBlockNumber = update.latestBlockNumber;
                        this.lastDepositIdForSpokePool = update.latestDepositId;
                        this.firstBlockToSearch = update.searchEndBlock + 1;
                        this.latestBlockSearched = update.searchEndBlock;
                        this.isUpdated = true;
                        this.log("debug", "SpokePool client for chain ".concat(this.chainId, " updated!"), {
                            nextFirstBlockToSearch: this.firstBlockToSearch,
                        });
                        return [2];
                }
            });
        });
    };
    SpokePoolClient.getExecutedRefundLeafL2Token = function (chainId, eventL2Token) {
        if ((chainId === 10 || chainId === 8453) &&
            eventL2Token.toLowerCase() === "0xdeaddeaddeaddeaddeaddeaddeaddeaddead0000") {
            return "0x4200000000000000000000000000000000000006";
        }
        else if (chainId === 288 && eventL2Token.toLowerCase() === "0x4200000000000000000000000000000000000006") {
            return "0xDeadDeAddeAddEAddeadDEaDDEAdDeaDDeAD0000";
        }
        else {
            return eventL2Token;
        }
    };
    SpokePoolClient.prototype.computeRealizedLpFeePct = function (depositEvent) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var deposit, l1Token;
            return tslib_1.__generator(this, function (_a) {
                if (this.hubPoolClient === null) {
                    return [2, { realizedLpFeePct: (0, utils_1.toBN)(0), quoteBlock: utils_1.MAX_BIG_INT.toNumber() }];
                }
                deposit = {
                    amount: depositEvent.args.amount,
                    originChainId: Number(depositEvent.args.originChainId),
                    destinationChainId: Number(depositEvent.args.destinationChainId),
                    originToken: depositEvent.args.originToken,
                    quoteTimestamp: depositEvent.args.quoteTimestamp,
                    blockNumber: depositEvent.blockNumber,
                };
                l1Token = this.hubPoolClient.getL1TokenForDeposit(deposit);
                return [2, this.hubPoolClient.computeRealizedLpFeePct(deposit, l1Token)];
            });
        });
    };
    SpokePoolClient.prototype.getDestinationTokenForDeposit = function (deposit) {
        if (!this.hubPoolClient) {
            return constants_1.ZERO_ADDRESS;
        }
        return this.hubPoolClient.getDestinationTokenForDeposit(deposit);
    };
    SpokePoolClient.prototype.log = function (level, message, data) {
        this.logger[level](tslib_1.__assign({ at: "SpokePoolClient", chainId: this.chainId, message: message }, data));
    };
    SpokePoolClient.prototype.getCurrentTime = function () {
        return this.currentTime;
    };
    SpokePoolClient.prototype.findDeposit = function (depositId, destinationChainId, depositor) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var searchBounds, tStart, query, tStop, event, srcChain, dstChain, partialDeposit, _a, realizedLpFeePct, quoteBlockNumber, deposit;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4, this._getBlockRangeForDepositId(depositId + 1, this.deploymentBlock, this.latestBlockNumber, 7)];
                    case 1:
                        searchBounds = _b.sent();
                        tStart = Date.now();
                        return [4, (0, EventUtils_1.paginatedEventQuery)(this.spokePool, this.spokePool.filters.FundsDeposited(null, null, destinationChainId, null, depositId, null, null, null, depositor, null), {
                                fromBlock: searchBounds.low,
                                toBlock: searchBounds.high,
                                maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                            })];
                    case 2:
                        query = _b.sent();
                        tStop = Date.now();
                        event = query.find(function (deposit) { return deposit.args.depositId === depositId; });
                        if (event === undefined) {
                            srcChain = (0, NetworkUtils_1.getNetworkName)(this.chainId);
                            dstChain = (0, NetworkUtils_1.getNetworkName)(destinationChainId);
                            throw new Error("Could not find deposit ".concat(depositId, " for ").concat(dstChain, " fill") +
                                " between ".concat(srcChain, " blocks [").concat(searchBounds.low, ", ").concat(searchBounds.high, "]"));
                        }
                        partialDeposit = (0, EventUtils_1.spreadEventWithBlockNumber)(event);
                        return [4, this.computeRealizedLpFeePct(event)];
                    case 3:
                        _a = _b.sent(), realizedLpFeePct = _a.realizedLpFeePct, quoteBlockNumber = _a.quoteBlock;
                        deposit = tslib_1.__assign(tslib_1.__assign({}, partialDeposit), { realizedLpFeePct: realizedLpFeePct, destinationToken: this.getDestinationTokenForDeposit(partialDeposit), quoteBlockNumber: quoteBlockNumber });
                        this.logger.debug({
                            at: "SpokePoolClient#findDeposit",
                            message: "Located deposit outside of SpokePoolClient's search range",
                            deposit: deposit,
                            elapsedMs: tStop - tStart,
                        });
                        return [2, deposit];
                }
            });
        });
    };
    SpokePoolClient.prototype.updateFromJSON = function (spokePoolClientState) {
        var keysToUpdate = Object.keys(spokePoolClientState);
        this.logger.debug({
            at: "SpokePoolClient",
            message: "Updating SpokePool client from JSON",
            keys: keysToUpdate,
        });
        if (keysToUpdate.length === 0) {
            return;
        }
        this.currentTime = spokePoolClientState.currentTime || this.currentTime;
        this.depositHashes = Object.entries(spokePoolClientState.depositHashes || {}).reduce(function (acc, _a) {
            var _b;
            var hash = _a[0], deposit = _a[1];
            return (tslib_1.__assign(tslib_1.__assign({}, acc), (_b = {}, _b[hash] = tslib_1.__assign(tslib_1.__assign({}, deposit), { amount: ethers_1.BigNumber.from(deposit.amount), relayerFeePct: ethers_1.BigNumber.from(deposit.relayerFeePct), realizedLpFeePct: deposit.relayerFeePct ? ethers_1.BigNumber.from(deposit.realizedLpFeePct) : undefined, newRelayerFeePct: deposit.newRelayerFeePct ? ethers_1.BigNumber.from(deposit.newRelayerFeePct) : undefined }), _b)));
        }, {});
        this.depositHashesToFills = Object.entries(spokePoolClientState.depositHashesToFills || {}).reduce(function (acc, _a) {
            var _b;
            var hash = _a[0], fills = _a[1];
            return (tslib_1.__assign(tslib_1.__assign({}, acc), (_b = {}, _b[hash] = fills.map(function (fill) { return (tslib_1.__assign(tslib_1.__assign({}, fill), { amount: ethers_1.BigNumber.from(fill.amount), totalFilledAmount: ethers_1.BigNumber.from(fill.totalFilledAmount), fillAmount: ethers_1.BigNumber.from(fill.fillAmount), relayerFeePct: ethers_1.BigNumber.from(fill.relayerFeePct), realizedLpFeePct: ethers_1.BigNumber.from(fill.realizedLpFeePct), updatableRelayData: tslib_1.__assign(tslib_1.__assign({}, fill.updatableRelayData), { relayerFeePct: ethers_1.BigNumber.from(fill.updatableRelayData.relayerFeePct), payoutAdjustmentPct: ethers_1.BigNumber.from(fill.updatableRelayData.payoutAdjustmentPct) }) })); }), _b)));
        }, {});
        this.speedUps = Object.entries(spokePoolClientState.speedUps || {}).reduce(function (acc, _a) {
            var _b;
            var depositor = _a[0], speedUpsMap = _a[1];
            var speedUps = Object.entries(speedUpsMap).reduce(function (acc, _a) {
                var _b;
                var depositId = _a[0], speedUps = _a[1];
                return (tslib_1.__assign(tslib_1.__assign({}, acc), (_b = {}, _b[depositId] = speedUps.map(function (speedUp) { return (tslib_1.__assign(tslib_1.__assign({}, speedUp), { newRelayerFeePct: ethers_1.BigNumber.from(speedUp.newRelayerFeePct) })); }), _b)));
            }, {});
            return tslib_1.__assign(tslib_1.__assign({}, acc), (_b = {}, _b[depositor] = speedUps, _b));
        }, {});
        this.depositRoutes = spokePoolClientState.depositRoutes || {};
        this.tokensBridged = (spokePoolClientState.tokensBridged || []).map(function (tokenBridged) { return (tslib_1.__assign(tslib_1.__assign({}, tokenBridged), { amountToReturn: ethers_1.BigNumber.from(tokenBridged.amountToReturn) })); });
        this.rootBundleRelays = spokePoolClientState.rootBundleRelays || [];
        this.relayerRefundExecutions = (spokePoolClientState.relayerRefundExecutions || []).map(function (refundExecution) { return (tslib_1.__assign(tslib_1.__assign({}, refundExecution), { amountToReturn: ethers_1.BigNumber.from(refundExecution.amountToReturn), refundAmounts: refundExecution.refundAmounts.map(function (refundAmount) { return ethers_1.BigNumber.from(refundAmount); }) })); });
        this.earlyDeposits = (spokePoolClientState.earlyDeposits || []).map(function (deposit) { return (tslib_1.__assign(tslib_1.__assign({}, deposit), { amount: ethers_1.BigNumber.from(deposit.amount), originChainId: ethers_1.BigNumber.from(deposit.originChainId), destinationChainId: ethers_1.BigNumber.from(deposit.destinationChainId), relayerFeePct: ethers_1.BigNumber.from(deposit.relayerFeePct) })); });
        this.queryableEventNames = spokePoolClientState.queryableEventNames || this.queryableEventNames;
        this.earliestDepositIdQueried = spokePoolClientState.earliestDepositIdQueried || this.earliestDepositIdQueried;
        this.latestDepositIdQueried = spokePoolClientState.latestDepositIdQueried || this.latestDepositIdQueried;
        this.firstBlockToSearch = spokePoolClientState.firstBlockToSearch || this.firstBlockToSearch;
        this.latestBlockSearched = spokePoolClientState.latestBlockSearched || this.latestBlockSearched;
        this.fills = Object.entries(spokePoolClientState.fills || []).reduce(function (acc, _a) {
            var _b;
            var chainId = _a[0], fills = _a[1];
            return (tslib_1.__assign(tslib_1.__assign({}, acc), (_b = {}, _b[chainId] = fills.map(function (fill) { return (tslib_1.__assign(tslib_1.__assign({}, fill), { amount: ethers_1.BigNumber.from(fill.amount), totalFilledAmount: ethers_1.BigNumber.from(fill.totalFilledAmount), fillAmount: ethers_1.BigNumber.from(fill.fillAmount), relayerFeePct: ethers_1.BigNumber.from(fill.relayerFeePct), realizedLpFeePct: ethers_1.BigNumber.from(fill.realizedLpFeePct), updatableRelayData: tslib_1.__assign(tslib_1.__assign({}, fill.updatableRelayData), { relayerFeePct: ethers_1.BigNumber.from(fill.updatableRelayData.relayerFeePct), payoutAdjustmentPct: ethers_1.BigNumber.from(fill.updatableRelayData.payoutAdjustmentPct) }) })); }), _b)));
        }, {});
        this.refundRequests = (spokePoolClientState.refundRequests || []).map(function (refundRequest) { return (tslib_1.__assign(tslib_1.__assign({}, refundRequest), { amount: ethers_1.BigNumber.from(refundRequest.amount), realizedLpFeePct: ethers_1.BigNumber.from(refundRequest.realizedLpFeePct), previousIdenticalRequests: ethers_1.BigNumber.from(refundRequest.previousIdenticalRequests), fillBlock: ethers_1.BigNumber.from(refundRequest.fillBlock) })); });
        this.isUpdated = true;
    };
    SpokePoolClient.prototype.toJSON = function () {
        return {
            chainId: this.chainId,
            deploymentBlock: this.deploymentBlock,
            eventSearchConfig: this.eventSearchConfig,
            currentTime: this.currentTime,
            depositHashes: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.depositHashes)),
            depositHashesToFills: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.depositHashesToFills)),
            speedUps: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.speedUps)),
            depositRoutes: this.depositRoutes,
            tokensBridged: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.tokensBridged)),
            rootBundleRelays: this.rootBundleRelays,
            relayerRefundExecutions: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.relayerRefundExecutions)),
            earlyDeposits: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.earlyDeposits)),
            queryableEventNames: this.queryableEventNames,
            earliestDepositIdQueried: this.earliestDepositIdQueried,
            earliestDepositIdForSpokePool: this.firstDepositIdForSpokePool,
            latestDepositIdQueried: this.latestDepositIdQueried,
            latestDepositIdForSpokePool: this.lastDepositIdForSpokePool,
            firstBlockToSearch: this.firstBlockToSearch,
            latestBlockSearched: this.latestBlockSearched,
            isUpdated: this.isUpdated,
            fills: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.fills)),
            refundRequests: JSON.parse((0, utils_1.stringifyJSONWithNumericString)(this.refundRequests)),
        };
    };
    return SpokePoolClient;
}(BaseAbstractClient_1.BaseAbstractClient));
exports.SpokePoolClient = SpokePoolClient;
//# sourceMappingURL=SpokePoolClient.js.map