"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RelayFeeCalculator = exports.DEFAULT_LOGGER = exports.expectedCapitalCostsKeys = void 0;
var tslib_1 = require("tslib");
var assert_1 = tslib_1.__importDefault(require("assert"));
var uma = tslib_1.__importStar(require("@uma/sdk"));
var utils_1 = require("../utils");
var _a = uma.across.utils, percent = _a.percent, fixedPointAdjustment = _a.fixedPointAdjustment;
exports.expectedCapitalCostsKeys = ["lowerBound", "upperBound", "cutoff", "decimals"];
exports.DEFAULT_LOGGER = {
    debug: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.debug(args);
    },
    info: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.info(args);
    },
    warn: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.warn(args);
    },
    error: function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return console.error(args);
    },
};
var RelayFeeCalculator = (function () {
    function RelayFeeCalculator(config, logger, destinationChainId) {
        (0, assert_1.default)(config, "config must be provided");
        if ("queries" in config) {
            this.queries = config.queries;
        }
        else {
            (0, assert_1.default)(destinationChainId !== undefined, "destinationChainId must be provided if queriesMap is provided");
            (0, assert_1.default)(config.queriesMap[destinationChainId], "No queries provided for destination chain");
            this.queries = config.queriesMap[destinationChainId];
        }
        this.gasDiscountPercent = config.gasDiscountPercent || 0;
        this.capitalDiscountPercent = config.capitalDiscountPercent || 0;
        this.feeLimitPercent = config.feeLimitPercent || 0;
        this.nativeTokenDecimals = config.nativeTokenDecimals || 18;
        this.capitalCostsPercent = config.capitalCostsPercent || 0;
        (0, assert_1.default)(this.gasDiscountPercent >= 0 && this.gasDiscountPercent <= 100, "gasDiscountPercent must be between 0 and 100 percent");
        (0, assert_1.default)(this.capitalDiscountPercent >= 0 && this.capitalDiscountPercent <= 100, "capitalDiscountPercent must be between 0 and 100 percent");
        (0, assert_1.default)(this.feeLimitPercent >= 0 && this.feeLimitPercent <= 100, "feeLimitPercent must be between 0 and 100 percent");
        (0, assert_1.default)(this.capitalCostsPercent >= 0 && this.capitalCostsPercent <= 100, "capitalCostsPercent must be between 0 and 100 percent");
        this.capitalCostsConfig = {};
        for (var _i = 0, _a = Object.keys(config.capitalCostsConfig || {}); _i < _a.length; _i++) {
            var token = _a[_i];
            this.capitalCostsConfig[token] = RelayFeeCalculator.validateAndTransformCapitalCostsConfigOverride((config.capitalCostsConfig || {})[token]);
        }
        this.logger = logger || exports.DEFAULT_LOGGER;
    }
    RelayFeeCalculator.capitalCostConfigIsOverride = function (config) {
        return config.default !== undefined;
    };
    RelayFeeCalculator.validateAndTransformCapitalCostsConfigOverride = function (capitalCosts) {
        var config = this.capitalCostConfigIsOverride(capitalCosts)
            ? capitalCosts
            : { default: capitalCosts };
        this.validateCapitalCostsConfig(config.default);
        for (var _i = 0, _a = Object.values(config.routeOverrides || {}); _i < _a.length; _i++) {
            var toChainIdRoutes = _a[_i];
            for (var _b = 0, _c = Object.values(toChainIdRoutes); _b < _c.length; _b++) {
                var override = _c[_b];
                this.validateCapitalCostsConfig(override);
            }
        }
        return config;
    };
    RelayFeeCalculator.validateCapitalCostsConfig = function (capitalCosts) {
        (0, assert_1.default)((0, utils_1.toBN)(capitalCosts.upperBound).lt((0, utils_1.toBNWei)("0.01")), "upper bound must be < 1%");
        (0, assert_1.default)((0, utils_1.toBN)(capitalCosts.lowerBound).lte(capitalCosts.upperBound), "lower bound must be <= upper bound");
        (0, assert_1.default)(capitalCosts.decimals > 0 && capitalCosts.decimals <= 18, "invalid decimals");
    };
    RelayFeeCalculator.prototype.getTokenPrice = function (tokenSymbol) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2, this.queries.getTokenPrice(tokenSymbol)];
            });
        });
    };
    RelayFeeCalculator.prototype.gasFeePercent = function (amountToRelay, tokenSymbol, _tokenPrice) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var getGasCosts, getTokenPrice, _a, gasCosts, tokenPrice, decimals, gasFeesInToken;
            var _this = this;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if ((0, utils_1.toBN)(amountToRelay).eq(0))
                            return [2, utils_1.MAX_BIG_INT];
                        getGasCosts = this.queries.getGasCosts().catch(function (error) {
                            _this.logger.error({ at: "sdk-v2/gasFeePercent", message: "Error while fetching gas costs", error: error });
                            throw error;
                        });
                        getTokenPrice = this.queries.getTokenPrice(tokenSymbol).catch(function (error) {
                            _this.logger.error({ at: "sdk-v2/gasFeePercent", message: "Error while fetching token price", error: error });
                            throw error;
                        });
                        return [4, Promise.all([
                                getGasCosts,
                                _tokenPrice !== undefined ? _tokenPrice : getTokenPrice,
                            ])];
                    case 1:
                        _a = _b.sent(), gasCosts = _a[0], tokenPrice = _a[1];
                        decimals = this.queries.getTokenDecimals(tokenSymbol);
                        gasFeesInToken = (0, utils_1.nativeToToken)(gasCosts, tokenPrice, decimals, this.nativeTokenDecimals);
                        return [2, percent(gasFeesInToken, amountToRelay.toString())];
                }
            });
        });
    };
    RelayFeeCalculator.prototype.capitalFeePercent = function (_amountToRelay, _tokenSymbol, _originRoute, _destinationRoute) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var defaultFee, overrideConfig, config, potentialDestinationRoutes, y, minCharge, yTriangle, triangleSlope, triangleHeight, triangleCharge, yRemainder, remainderCharge;
            return tslib_1.__generator(this, function (_a) {
                if ((0, utils_1.toBN)(_amountToRelay).eq((0, utils_1.toBN)(0)))
                    return [2, utils_1.MAX_BIG_INT];
                defaultFee = (0, utils_1.toBNWei)(this.capitalCostsPercent / 100);
                if (this.capitalCostsConfig[_tokenSymbol]) {
                    overrideConfig = this.capitalCostsConfig[_tokenSymbol];
                    config = overrideConfig.default;
                    if (_originRoute && _destinationRoute && overrideConfig.routeOverrides) {
                        potentialDestinationRoutes = overrideConfig.routeOverrides[_originRoute];
                        if (potentialDestinationRoutes) {
                            config = potentialDestinationRoutes[_destinationRoute] || config;
                        }
                    }
                    y = (0, utils_1.toBN)(_amountToRelay).mul((0, utils_1.toBNWei)("1", 18 - config.decimals));
                    minCharge = (0, utils_1.toBN)(config.lowerBound).mul(y).div(fixedPointAdjustment);
                    yTriangle = (0, utils_1.min)(config.cutoff, y);
                    triangleSlope = (0, utils_1.toBN)(config.cutoff).eq((0, utils_1.toBN)(0))
                        ? (0, utils_1.toBN)(0)
                        : (0, utils_1.toBN)(config.upperBound).sub(config.lowerBound).mul(fixedPointAdjustment).div(config.cutoff);
                    triangleHeight = triangleSlope.mul(yTriangle).div(fixedPointAdjustment);
                    triangleCharge = triangleHeight.mul(yTriangle).div((0, utils_1.toBNWei)(2));
                    yRemainder = (0, utils_1.max)((0, utils_1.toBN)(0), y.sub(config.cutoff));
                    remainderCharge = yRemainder.mul((0, utils_1.toBN)(config.upperBound).sub(config.lowerBound)).div(fixedPointAdjustment);
                    return [2, minCharge.add(triangleCharge).add(remainderCharge).mul(fixedPointAdjustment).div(y)];
                }
                return [2, defaultFee];
            });
        });
    };
    RelayFeeCalculator.prototype.relayerFeeDetails = function (amountToRelay, tokenSymbol, tokenPrice, _originRoute, _destinationRoute) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var gasFeePercent, gasFeeTotal, capitalFeePercent, capitalFeeTotal, relayFeePercent, relayFeeTotal, maxGasFeePercent, minDeposit, isAmountTooLow;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.gasFeePercent(amountToRelay, tokenSymbol, tokenPrice)];
                    case 1:
                        gasFeePercent = _a.sent();
                        gasFeeTotal = gasFeePercent.mul(amountToRelay).div(fixedPointAdjustment);
                        return [4, this.capitalFeePercent(amountToRelay, tokenSymbol, _originRoute, _destinationRoute)];
                    case 2:
                        capitalFeePercent = _a.sent();
                        capitalFeeTotal = capitalFeePercent.mul(amountToRelay).div(fixedPointAdjustment);
                        relayFeePercent = gasFeePercent.add(capitalFeePercent);
                        relayFeeTotal = gasFeeTotal.add(capitalFeeTotal);
                        maxGasFeePercent = (0, utils_1.max)((0, utils_1.toBNWei)(this.feeLimitPercent / 100).sub(capitalFeePercent), (0, utils_1.toBN)(0));
                        if (maxGasFeePercent.eq((0, utils_1.toBN)(0))) {
                            minDeposit = utils_1.MAX_BIG_INT;
                            isAmountTooLow = true;
                        }
                        else {
                            minDeposit = gasFeeTotal.mul(fixedPointAdjustment).div(maxGasFeePercent);
                            isAmountTooLow = (0, utils_1.toBN)(amountToRelay).lt(minDeposit);
                        }
                        return [2, {
                                amountToRelay: amountToRelay.toString(),
                                tokenSymbol: tokenSymbol,
                                gasFeePercent: gasFeePercent.toString(),
                                gasFeeTotal: gasFeeTotal.toString(),
                                gasDiscountPercent: this.gasDiscountPercent,
                                capitalFeePercent: capitalFeePercent.toString(),
                                capitalFeeTotal: capitalFeeTotal.toString(),
                                capitalDiscountPercent: this.capitalDiscountPercent,
                                relayFeePercent: relayFeePercent.toString(),
                                relayFeeTotal: relayFeeTotal.toString(),
                                feeLimitPercent: this.feeLimitPercent,
                                maxGasFeePercent: maxGasFeePercent.toString(),
                                minDeposit: minDeposit.toString(),
                                isAmountTooLow: isAmountTooLow,
                            }];
                }
            });
        });
    };
    return RelayFeeCalculator;
}());
exports.RelayFeeCalculator = RelayFeeCalculator;
//# sourceMappingURL=relayFeeCalculator.js.map