import { __assign, __awaiter, __extends, __generator, __spreadArray } from "tslib";
import { BigNumber } from "ethers";
import { BlockFinder } from "@uma/sdk";
import _ from "lodash";
import { assign, stringifyJSONWithNumericString, isDefined, getCurrentTime, shouldCache, } from "../utils";
import { fetchTokenInfo, sortEventsDescending, spreadEvent, spreadEventWithBlockNumber, paginatedEventQuery, toBN, } from "../utils";
import * as lpFeeCalculator from "../lpFeeCalculator";
import { BaseAbstractClient } from "./BaseAbstractClient";
import { isUBAActivatedAtBlock } from "./UBAClient/UBAClientUtilities";
import { DEFAULT_CACHING_TTL } from "../constants";
var HubPoolClient = /** @class */ (function (_super) {
    __extends(HubPoolClient, _super);
    function HubPoolClient(logger, hubPool, configStoreClient, deploymentBlock, chainId, eventSearchConfig, configOverride, cachingMechanism) {
        if (deploymentBlock === void 0) { deploymentBlock = 0; }
        if (chainId === void 0) { chainId = 1; }
        if (eventSearchConfig === void 0) { eventSearchConfig = { fromBlock: 0, maxBlockLookBack: 0 }; }
        if (configOverride === void 0) { configOverride = {
            ignoredHubExecutedBundles: [],
            ignoredHubProposedBundles: [],
        }; }
        var _this = _super.call(this, cachingMechanism) || this;
        _this.logger = logger;
        _this.hubPool = hubPool;
        _this.configStoreClient = configStoreClient;
        _this.deploymentBlock = deploymentBlock;
        _this.chainId = chainId;
        _this.eventSearchConfig = eventSearchConfig;
        _this.configOverride = configOverride;
        // L1Token -> destinationChainId -> destinationToken
        _this.l1TokensToDestinationTokens = {};
        _this.l1Tokens = []; // L1Tokens and their associated info.
        _this.lpTokens = {};
        _this.proposedRootBundles = [];
        _this.canceledRootBundles = [];
        _this.disputedRootBundles = [];
        _this.executedRootBundles = [];
        _this.crossChainContracts = {};
        _this.l1TokensToDestinationTokensWithBlock = {};
        _this.latestBlockNumber = deploymentBlock === 0 ? deploymentBlock : deploymentBlock - 1;
        _this.firstBlockToSearch = eventSearchConfig.fromBlock;
        var provider = _this.hubPool.provider;
        _this.blockFinder = new BlockFinder(provider.getBlock.bind(provider));
        return _this;
    }
    HubPoolClient.prototype.hubPoolEventFilters = function () {
        return {
            SetPoolRebalanceRoute: this.hubPool.filters.SetPoolRebalanceRoute(),
            L1TokenEnabledForLiquidityProvision: this.hubPool.filters.L1TokenEnabledForLiquidityProvision(),
            ProposeRootBundle: this.hubPool.filters.ProposeRootBundle(),
            RootBundleCanceled: this.hubPool.filters.RootBundleCanceled(),
            RootBundleDisputed: this.hubPool.filters.RootBundleDisputed(),
            RootBundleExecuted: this.hubPool.filters.RootBundleExecuted(),
            CrossChainContractsSet: this.hubPool.filters.CrossChainContractsSet(),
        };
    };
    HubPoolClient.prototype.hasPendingProposal = function () {
        return this.pendingRootBundle !== undefined;
    };
    HubPoolClient.prototype.getPendingRootBundle = function () {
        return this.pendingRootBundle;
    };
    HubPoolClient.prototype.getProposedRootBundles = function () {
        return this.proposedRootBundles;
    };
    HubPoolClient.prototype.getCancelledRootBundles = function () {
        return this.canceledRootBundles;
    };
    HubPoolClient.prototype.getDisputedRootBundles = function () {
        return this.disputedRootBundles;
    };
    HubPoolClient.prototype.getSpokePoolForBlock = function (chain, block) {
        if (block === void 0) { block = Number.MAX_SAFE_INTEGER; }
        if (!this.crossChainContracts[chain]) {
            throw new Error("No cross chain contracts set for ".concat(chain));
        }
        var mostRecentSpokePoolUpdateBeforeBlock = sortEventsDescending(this.crossChainContracts[chain]).find(function (crossChainContract) { return crossChainContract.blockNumber <= block; });
        if (!mostRecentSpokePoolUpdateBeforeBlock) {
            throw new Error("No cross chain contract found before block ".concat(block, " for chain ").concat(chain));
        }
        else {
            return mostRecentSpokePoolUpdateBeforeBlock.spokePool;
        }
    };
    HubPoolClient.prototype.getSpokePoolActivationBlock = function (chain, spokePool) {
        // Return first time that this spoke pool was registered in the HubPool as a cross chain contract. We can use
        // this block as the oldest block that we should query for SpokePoolClient purposes.
        var mostRecentSpokePoolUpdateBeforeBlock = this.crossChainContracts[chain].find(function (crossChainContract) { return crossChainContract.spokePool === spokePool; });
        return mostRecentSpokePoolUpdateBeforeBlock === null || mostRecentSpokePoolUpdateBeforeBlock === void 0 ? void 0 : mostRecentSpokePoolUpdateBeforeBlock.blockNumber;
    };
    HubPoolClient.prototype.getDestinationTokenForDeposit = function (deposit) {
        var l1Token = this.getL1TokenForDeposit(deposit);
        var destinationToken = this.getDestinationTokenForL1Token(l1Token, deposit.destinationChainId);
        if (!destinationToken) {
            this.logger.error({
                at: "HubPoolClient",
                message: "No destination token found",
                deposit: deposit,
                notificationPath: "across-error",
            });
        }
        return destinationToken;
    };
    HubPoolClient.prototype.getL1TokensToDestinationTokens = function () {
        return this.l1TokensToDestinationTokens;
    };
    HubPoolClient.prototype.getL1TokenForDeposit = function (deposit) {
        var _this = this;
        var l1Token = Object.keys(this.l1TokensToDestinationTokens).find(function (_l1Token) {
            return _this.l1TokensToDestinationTokens[_l1Token][deposit.originChainId] === deposit.originToken;
        });
        if (!l1Token) {
            throw new Error("Could not find L1 Token for origin chain ".concat(deposit.originChainId, " and origin token ").concat(deposit.originToken, "!"));
        }
        return l1Token;
    };
    HubPoolClient.prototype.getL1TokenCounterpartAtBlock = function (l2ChainId, l2Token, hubPoolBlock) {
        var _this = this;
        var l1Token = Object.keys(this.l1TokensToDestinationTokensWithBlock).find(function (_l1Token) {
            // If this token doesn't exist on this L2, return false.
            if (_this.l1TokensToDestinationTokensWithBlock[_l1Token][l2ChainId] === undefined) {
                return false;
            }
            // Find the last mapping published before the target block.
            return sortEventsDescending(_this.l1TokensToDestinationTokensWithBlock[_l1Token][l2ChainId]).find(function (mapping) { return mapping.l2Token === l2Token && mapping.blockNumber <= hubPoolBlock; });
        });
        if (!l1Token) {
            throw new Error("Could not find L1 token mapping for chain ".concat(l2ChainId, " and L2 token ").concat(l2Token, " equal to or earlier than block ").concat(hubPoolBlock, "!"));
        }
        return l1Token;
    };
    HubPoolClient.prototype.getDestinationTokenForL1Token = function (l1Token, destinationChainId) {
        return this.l1TokensToDestinationTokens[l1Token][destinationChainId];
    };
    HubPoolClient.prototype.l2TokenEnabledForL1Token = function (l1Token, destinationChainId) {
        return this.l1TokensToDestinationTokens[l1Token][destinationChainId] != undefined;
    };
    HubPoolClient.prototype.getDestinationTokensToL1TokensForChainId = function (chainId) {
        var _this = this;
        return Object.fromEntries(this.l1Tokens
            .map(function (l1Token) { return [_this.getDestinationTokenForL1Token(l1Token.address, chainId), l1Token]; })
            .filter(function (entry) { return entry[0] !== undefined; }));
    };
    HubPoolClient.prototype.getBlockNumber = function (timestamp) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.blockFinder.getBlockForTimestamp(timestamp)];
                    case 1: return [2 /*return*/, (_a.sent()).number];
                }
            });
        });
    };
    HubPoolClient.prototype.getCurrentPoolUtilization = function (l1Token) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.hubPool.callStatic.liquidityUtilizationCurrent(l1Token)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    HubPoolClient.prototype.getPostRelayPoolUtilization = function (l1Token, quoteBlockNumber, relaySize) {
        return __awaiter(this, void 0, void 0, function () {
            var overrides, _a, current, post;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        overrides = { blockTag: quoteBlockNumber };
                        return [4 /*yield*/, Promise.all([
                                this.hubPool.callStatic.liquidityUtilizationCurrent(l1Token, overrides),
                                this.hubPool.callStatic.liquidityUtilizationPostRelay(l1Token, relaySize, overrides),
                            ])];
                    case 1:
                        _a = _b.sent(), current = _a[0], post = _a[1];
                        return [2 /*return*/, { current: current, post: post }];
                }
            });
        });
    };
    HubPoolClient.prototype.getUtilization = function (l1Token, blockNumber, amount, timestamp) {
        return __awaiter(this, void 0, void 0, function () {
            var resolver, cache, key, result, _a, current, post, _b, current, post;
            var _this = this;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        resolver = function () { return __awaiter(_this, void 0, void 0, function () { return __generator(this, function (_a) {
                            return [2 /*return*/, this.getPostRelayPoolUtilization(l1Token, blockNumber, amount)];
                        }); }); };
                        cache = this.cachingMechanism;
                        // If there is no cache, just resolve the function
                        if (!cache) {
                            return [2 /*return*/, resolver()];
                        }
                        key = "utilization_".concat(l1Token, "_").concat(blockNumber, "_").concat(amount.toString());
                        return [4 /*yield*/, cache.get(key)];
                    case 1:
                        result = _c.sent();
                        if (!isDefined(result)) return [3 /*break*/, 2];
                        _a = result.split(",").map(BigNumber.from), current = _a[0], post = _a[1];
                        return [2 /*return*/, { current: current, post: post }];
                    case 2: return [4 /*yield*/, resolver()];
                    case 3:
                        _b = _c.sent(), current = _b.current, post = _b.post;
                        if (!shouldCache(getCurrentTime(), timestamp, 60 * 60 * 24)) return [3 /*break*/, 5];
                        // If we should cache the result, then let's store it
                        // We can store it as with the default 14 day TTL
                        return [4 /*yield*/, cache.set(key, "".concat(current.toString(), ",").concat(post.toString()), DEFAULT_CACHING_TTL)];
                    case 4:
                        // If we should cache the result, then let's store it
                        // We can store it as with the default 14 day TTL
                        _c.sent();
                        _c.label = 5;
                    case 5: 
                    // Return the result
                    return [2 /*return*/, { current: current, post: post }];
                }
            });
        });
    };
    HubPoolClient.prototype.computeRealizedLpFeePct = function (deposit, l1Token) {
        return __awaiter(this, void 0, void 0, function () {
            var quoteBlock, rateModel, _a, current, post, realizedLpFeePct;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if (!isDefined(this.currentTime)) {
                            throw new Error("HubPoolClient has not set a currentTime");
                        }
                        return [4 /*yield*/, this.getBlockNumber(deposit.quoteTimestamp)];
                    case 1:
                        quoteBlock = _b.sent();
                        if (!isDefined(quoteBlock)) {
                            throw new Error("Could not find block for timestamp ".concat(deposit.quoteTimestamp));
                        }
                        // Compare deposit block against UBA bundle start blocks.
                        if (isUBAActivatedAtBlock(this, deposit.blockNumber, deposit.originChainId)) {
                            // If UBA deposit then we can't compute the realizedLpFeePct until after we've updated the UBA Client.
                            return [2 /*return*/, {
                                    realizedLpFeePct: undefined,
                                    quoteBlock: quoteBlock,
                                }];
                        }
                        rateModel = this.configStoreClient.getRateModelForBlockNumber(l1Token, deposit.originChainId, deposit.destinationChainId, quoteBlock);
                        return [4 /*yield*/, this.getUtilization(l1Token, quoteBlock, deposit.amount, deposit.quoteTimestamp)];
                    case 2:
                        _a = _b.sent(), current = _a.current, post = _a.post;
                        realizedLpFeePct = lpFeeCalculator.calculateRealizedLpFeePct(rateModel, current, post);
                        return [2 /*return*/, { realizedLpFeePct: realizedLpFeePct, quoteBlock: quoteBlock }];
                }
            });
        });
    };
    HubPoolClient.prototype.getL1Tokens = function () {
        return this.l1Tokens;
    };
    HubPoolClient.prototype.getTokenInfoForL1Token = function (l1Token) {
        return this.l1Tokens.find(function (token) { return token.address === l1Token; });
    };
    HubPoolClient.prototype.getLpTokenInfoForL1Token = function (l1Token) {
        return this.lpTokens[l1Token];
    };
    HubPoolClient.prototype.getL1TokenInfoForL2Token = function (l2Token, chainId) {
        var l1TokenCounterpart = this.getL1TokenCounterpartAtBlock(chainId, l2Token, this.latestBlockNumber || 0);
        return this.getTokenInfoForL1Token(l1TokenCounterpart);
    };
    HubPoolClient.prototype.getTokenInfoForDeposit = function (deposit) {
        return this.getTokenInfoForL1Token(this.getL1TokenForDeposit(deposit));
    };
    HubPoolClient.prototype.getTokenInfo = function (chainId, tokenAddress) {
        var deposit = { originChainId: parseInt(chainId.toString()), originToken: tokenAddress };
        return this.getTokenInfoForDeposit(deposit);
    };
    HubPoolClient.prototype.getSpokeActivationBlockForChain = function (chainId) {
        var _a;
        return (_a = this.getSpokePoolActivationBlock(chainId, this.getSpokePoolForBlock(chainId))) !== null && _a !== void 0 ? _a : 0;
    };
    // Root bundles are valid if all of their pool rebalance leaves have been executed before the next bundle, or the
    // latest mainnet block to search. Whichever comes first.
    HubPoolClient.prototype.isRootBundleValid = function (rootBundle, latestMainnetBlock) {
        var nextRootBundle = this.getFollowingRootBundle(rootBundle);
        var executedLeafCount = this.getExecutedLeavesForRootBundle(rootBundle, nextRootBundle ? Math.min(nextRootBundle.blockNumber, latestMainnetBlock) : latestMainnetBlock);
        return executedLeafCount.length === rootBundle.poolRebalanceLeafCount;
    };
    // This should find the ProposeRootBundle event whose bundle block number for `chain` is closest to the `block`
    // without being smaller. It returns the bundle block number for the chain or undefined if not matched.
    HubPoolClient.prototype.getRootBundleEvalBlockNumberContainingBlock = function (latestMainnetBlock, block, chain, chainIdListOverride) {
        var chainIdList = chainIdListOverride !== null && chainIdListOverride !== void 0 ? chainIdListOverride : this.configStoreClient.getChainIdIndicesForBlock(latestMainnetBlock);
        var endingBlockNumber;
        // Search proposed root bundles in reverse chronological order.
        for (var i = this.proposedRootBundles.length - 1; i >= 0; i--) {
            var rootBundle = this.proposedRootBundles[i];
            var nextRootBundle = this.getFollowingRootBundle(rootBundle);
            if (!this.isRootBundleValid(rootBundle, nextRootBundle ? nextRootBundle.blockNumber : latestMainnetBlock)) {
                continue;
            }
            // 0 is the default value bundleEvalBlockNumber.
            var bundleEvalBlockNumber = this.getBundleEndBlockForChain(rootBundle, chain, chainIdList);
            // Since we're iterating from newest to oldest, bundleEvalBlockNumber is only decreasing, and if the
            // bundleEvalBlockNumber is smaller than the target block, then we should return the last set `endingBlockNumber`.
            if (bundleEvalBlockNumber <= block) {
                if (bundleEvalBlockNumber === block) {
                    endingBlockNumber = bundleEvalBlockNumber;
                }
                break;
            }
            endingBlockNumber = bundleEvalBlockNumber;
        }
        return endingBlockNumber;
    };
    // TODO: This might not be necessary since the cumulative root bundle count doesn't grow fast enough, but consider
    // using _.findLast/_.find instead of resorting the arrays if these functions begin to take a lot time.
    HubPoolClient.prototype.getProposedRootBundlesInBlockRange = function (startingBlock, endingBlock) {
        return this.proposedRootBundles.filter(function (bundle) { return bundle.blockNumber >= startingBlock && bundle.blockNumber <= endingBlock; });
    };
    HubPoolClient.prototype.getCancelledRootBundlesInBlockRange = function (startingBlock, endingBlock) {
        return sortEventsDescending(this.canceledRootBundles).filter(function (bundle) { return bundle.blockNumber >= startingBlock && bundle.blockNumber <= endingBlock; });
    };
    HubPoolClient.prototype.getDisputedRootBundlesInBlockRange = function (startingBlock, endingBlock) {
        return sortEventsDescending(this.disputedRootBundles).filter(function (bundle) { return bundle.blockNumber >= startingBlock && bundle.blockNumber <= endingBlock; });
    };
    HubPoolClient.prototype.getLatestProposedRootBundle = function () {
        return this.proposedRootBundles[this.proposedRootBundles.length - 1];
    };
    HubPoolClient.prototype.getFollowingRootBundle = function (currentRootBundle) {
        var index = _.findLastIndex(this.proposedRootBundles, function (bundle) { return bundle.blockNumber === currentRootBundle.blockNumber; });
        // If index of current root bundle is not found or is the last bundle, return undefined.
        if (index === -1 || index === this.proposedRootBundles.length - 1) {
            return undefined;
        }
        return this.proposedRootBundles[index + 1];
    };
    HubPoolClient.prototype.getExecutedLeavesForRootBundle = function (rootBundle, latestMainnetBlockToSearch) {
        return this.executedRootBundles.filter(function (executedLeaf) {
            return executedLeaf.blockNumber <= latestMainnetBlockToSearch &&
                // Note: We can use > instead of >= here because a leaf can never be executed in same block as its root
                // proposal due to bundle liveness enforced by HubPool. This importantly avoids the edge case
                // where the execution all leaves occurs in the same block as the next proposal, leading us to think
                // that the next proposal is fully executed when its not.
                executedLeaf.blockNumber > rootBundle.blockNumber;
        });
    };
    HubPoolClient.prototype.getValidatedRootBundles = function (latestMainnetBlock) {
        var _this = this;
        if (latestMainnetBlock === void 0) { latestMainnetBlock = Number.MAX_SAFE_INTEGER; }
        return this.proposedRootBundles.filter(function (rootBundle) {
            if (rootBundle.blockNumber > latestMainnetBlock) {
                return false;
            }
            return _this.isRootBundleValid(rootBundle, latestMainnetBlock);
        });
    };
    HubPoolClient.prototype.getLatestFullyExecutedRootBundle = function (latestMainnetBlock) {
        var _this = this;
        // Search for latest ProposeRootBundleExecuted event followed by all of its RootBundleExecuted event suggesting
        // that all pool rebalance leaves were executed. This ignores any proposed bundles that were partially executed.
        return _.findLast(this.proposedRootBundles, function (rootBundle) {
            if (rootBundle.blockNumber > latestMainnetBlock) {
                return false;
            }
            return _this.isRootBundleValid(rootBundle, latestMainnetBlock);
        });
    };
    HubPoolClient.prototype.getEarliestFullyExecutedRootBundle = function (latestMainnetBlock, startBlock) {
        var _this = this;
        if (startBlock === void 0) { startBlock = 0; }
        return this.proposedRootBundles.find(function (rootBundle) {
            if (rootBundle.blockNumber > latestMainnetBlock) {
                return false;
            }
            if (rootBundle.blockNumber < startBlock) {
                return false;
            }
            return _this.isRootBundleValid(rootBundle, latestMainnetBlock);
        });
    };
    // If n is negative, then return the Nth latest executed bundle, otherwise return the Nth earliest
    // executed bundle. Latest means most recent, earliest means oldest. N cannot be 0.
    // `startBlock` can be used to set the starting point from which we look forwards or backwards, depending
    // on whether n is positive or negative.
    HubPoolClient.prototype.getNthFullyExecutedRootBundle = function (n, startBlock) {
        if (n === 0) {
            throw new Error("n cannot be 0");
        }
        if (!this.latestBlockNumber) {
            throw new Error("HubPoolClient::getNthFullyExecutedRootBundle client not updated");
        }
        var bundleToReturn;
        // If n is negative, then return the Nth latest executed bundle, otherwise return the Nth earliest
        // executed bundle.
        if (n < 0) {
            var nextLatestMainnetBlock = startBlock !== null && startBlock !== void 0 ? startBlock : this.latestBlockNumber;
            for (var i = 0; i < Math.abs(n); i++) {
                bundleToReturn = this.getLatestFullyExecutedRootBundle(nextLatestMainnetBlock);
                var bundleBlockNumber = bundleToReturn ? bundleToReturn.blockNumber : 0;
                // Subtract 1 so that next `getLatestFullyExecutedRootBundle` call filters out the root bundle we just found
                // because its block number is > nextLatestMainnetBlock.
                nextLatestMainnetBlock = Math.max(0, bundleBlockNumber - 1);
            }
        }
        else {
            var nextStartBlock = startBlock !== null && startBlock !== void 0 ? startBlock : 0;
            for (var i = 0; i < n; i++) {
                bundleToReturn = this.getEarliestFullyExecutedRootBundle(this.latestBlockNumber, nextStartBlock);
                var bundleBlockNumber = bundleToReturn ? bundleToReturn.blockNumber : 0;
                // Add 1 so that next `getEarliestFullyExecutedRootBundle` call filters out the root bundle we just found
                // because its block number is < nextStartBlock.
                nextStartBlock = Math.min(bundleBlockNumber + 1, this.latestBlockNumber);
            }
        }
        return bundleToReturn;
    };
    HubPoolClient.prototype.getLatestBundleEndBlockForChain = function (chainIdList, latestMainnetBlock, chainId) {
        var latestFullyExecutedPoolRebalanceRoot = this.getLatestFullyExecutedRootBundle(latestMainnetBlock);
        // If no event, then we can return a conservative default starting block like 0,
        // or we could throw an Error.
        if (!latestFullyExecutedPoolRebalanceRoot) {
            return 0;
        }
        // Once this proposal event is found, determine its mapping of indices to chainId in its
        // bundleEvaluationBlockNumbers array using CHAIN_ID_LIST. For each chainId, their starting block number is that
        // chain's bundleEvaluationBlockNumber + 1 in this past proposal event.
        return this.getBundleEndBlockForChain(latestFullyExecutedPoolRebalanceRoot, chainId, chainIdList);
    };
    HubPoolClient.prototype.getNextBundleStartBlockNumber = function (chainIdList, latestMainnetBlock, chainId) {
        var endBlock = this.getLatestBundleEndBlockForChain(chainIdList, latestMainnetBlock, chainId);
        // This assumes that chain ID's are only added to the chain ID list over time, and that chains are never
        // deleted.
        return endBlock > 0 ? endBlock + 1 : 0;
    };
    HubPoolClient.prototype.getRunningBalanceBeforeBlockForChain = function (block, chain, l1Token) {
        // Search ExecutedRootBundles in descending block order to find the most recent event before the target block.
        var executedRootBundle = sortEventsDescending(this.executedRootBundles).find(function (executedLeaf) {
            return (executedLeaf.blockNumber <= block &&
                executedLeaf.chainId === chain &&
                executedLeaf.l1Tokens.map(function (l1Token) { return l1Token.toLowerCase(); }).includes(l1Token.toLowerCase()));
        });
        return this.getRunningBalanceForToken(l1Token, executedRootBundle);
    };
    HubPoolClient.prototype.getRunningBalanceForToken = function (l1Token, executedRootBundle) {
        var runningBalance = toBN(0);
        var incentiveBalance = toBN(0);
        if (executedRootBundle) {
            var indexOfL1Token = executedRootBundle.l1Tokens
                .map(function (l1Token) { return l1Token.toLowerCase(); })
                .indexOf(l1Token.toLowerCase());
            runningBalance = executedRootBundle.runningBalances[indexOfL1Token];
            incentiveBalance = executedRootBundle.incentiveBalances[indexOfL1Token];
        }
        return { runningBalance: runningBalance, incentiveBalance: incentiveBalance };
    };
    HubPoolClient.prototype._update = function (eventNames) {
        return __awaiter(this, void 0, void 0, function () {
            var latestBlockNumber, hubPoolEvents, searchConfig, timerStart, _a, currentTime, pendingRootBundleProposal, events, _events;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this.hubPool.provider.getBlockNumber()];
                    case 1:
                        latestBlockNumber = _b.sent();
                        hubPoolEvents = this.hubPoolEventFilters();
                        searchConfig = {
                            fromBlock: this.firstBlockToSearch,
                            toBlock: this.eventSearchConfig.toBlock || latestBlockNumber,
                            maxBlockLookBack: this.eventSearchConfig.maxBlockLookBack,
                        };
                        if (searchConfig.fromBlock > searchConfig.toBlock) {
                            this.logger.warn({ at: "HubPoolClient#_update", message: "Invalid update() searchConfig.", searchConfig: searchConfig });
                            return [2 /*return*/, { success: false }];
                        }
                        this.logger.debug({
                            at: "HubPoolClient",
                            message: "Updating HubPool client",
                            searchConfig: searchConfig,
                            eventNames: eventNames,
                        });
                        timerStart = Date.now();
                        return [4 /*yield*/, Promise.all(__spreadArray([
                                this.hubPool.getCurrentTime({ blockTag: searchConfig.toBlock }),
                                this.hubPool.rootBundleProposal({ blockTag: searchConfig.toBlock })
                            ], eventNames.map(function (eventName) { return paginatedEventQuery(_this.hubPool, hubPoolEvents[eventName], searchConfig); }), true))];
                    case 2:
                        _a = _b.sent(), currentTime = _a[0], pendingRootBundleProposal = _a[1], events = _a.slice(2);
                        this.logger.debug({
                            at: "HubPoolClient#_update",
                            message: "Time to query new events from RPC for ".concat(this.chainId, ": ").concat(Date.now() - timerStart, " ms"),
                        });
                        _events = Object.fromEntries(eventNames.map(function (eventName, idx) { return [eventName, events[idx]]; }));
                        return [2 /*return*/, {
                                success: true,
                                currentTime: currentTime,
                                latestBlockNumber: latestBlockNumber,
                                pendingRootBundleProposal: pendingRootBundleProposal,
                                searchEndBlock: searchConfig.toBlock,
                                events: _events,
                            }];
                }
            });
        });
    };
    HubPoolClient.prototype.update = function (eventsToQuery) {
        return __awaiter(this, void 0, void 0, function () {
            var update, events, currentTime, latestBlockNumber, pendingRootBundleProposal, _i, _a, event_1, args, _b, _c, event_2, args, uniqueL1Tokens, _d, tokenInfo, lpTokenInfo, _loop_1, this_1, _e, tokenInfo_1, info, _f, _g, event_3, executedRootBundle, l1Tokens, runningBalances, nTokens, mostRecentProposedRootBundle;
            var _h, _j, _k;
            var _this = this;
            return __generator(this, function (_l) {
                switch (_l.label) {
                    case 0:
                        if (!this.configStoreClient.isUpdated) {
                            throw new Error("ConfigStoreClient not updated");
                        }
                        eventsToQuery = eventsToQuery !== null && eventsToQuery !== void 0 ? eventsToQuery : Object.keys(this.hubPoolEventFilters()); // Query all events by default.
                        return [4 /*yield*/, this._update(eventsToQuery)];
                    case 1:
                        update = _l.sent();
                        if (!update.success) {
                            // This failure only occurs if the RPC searchConfig is miscomputed, and has only been seen in the hardhat test
                            // environment. Normal failures will throw instead. This is therefore an unfortunate workaround until we can
                            // understand why we see this in test. @todo: Resolve.
                            return [2 /*return*/];
                        }
                        events = update.events, currentTime = update.currentTime, latestBlockNumber = update.latestBlockNumber, pendingRootBundleProposal = update.pendingRootBundleProposal;
                        for (_i = 0, _a = events["CrossChainContractsSet"]; _i < _a.length; _i++) {
                            event_1 = _a[_i];
                            args = spreadEventWithBlockNumber(event_1);
                            assign(this.crossChainContracts, [args.l2ChainId], [
                                {
                                    spokePool: args.spokePool,
                                    blockNumber: args.blockNumber,
                                    transactionIndex: args.transactionIndex,
                                    logIndex: args.logIndex,
                                },
                            ]);
                        }
                        for (_b = 0, _c = events["SetPoolRebalanceRoute"]; _b < _c.length; _b++) {
                            event_2 = _c[_b];
                            args = spreadEventWithBlockNumber(event_2);
                            assign(this.l1TokensToDestinationTokens, [args.l1Token, args.destinationChainId], args.destinationToken);
                            assign(this.l1TokensToDestinationTokensWithBlock, [args.l1Token, args.destinationChainId], [
                                {
                                    l2Token: args.destinationToken,
                                    blockNumber: args.blockNumber,
                                    transactionIndex: args.transactionIndex,
                                    logIndex: args.logIndex,
                                },
                            ]);
                        }
                        uniqueL1Tokens = __spreadArray([], Array.from(new Set(events["L1TokenEnabledForLiquidityProvision"].map(function (event) { return spreadEvent(event.args).l1Token; }))), true);
                        return [4 /*yield*/, Promise.all([
                                Promise.all(uniqueL1Tokens.map(function (l1Token) { return fetchTokenInfo(l1Token, _this.hubPool.provider); })),
                                Promise.all(uniqueL1Tokens.map(function (l1Token) { return __awaiter(_this, void 0, void 0, function () { return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, this.hubPool.pooledTokens(l1Token, { blockTag: update.searchEndBlock })];
                                        case 1: return [2 /*return*/, _a.sent()];
                                    }
                                }); }); })),
                            ])];
                    case 2:
                        _d = _l.sent(), tokenInfo = _d[0], lpTokenInfo = _d[1];
                        _loop_1 = function (info) {
                            if (!this_1.l1Tokens.find(function (token) { return token.symbol === info.symbol; })) {
                                if (info.decimals > 0 && info.decimals <= 18) {
                                    this_1.l1Tokens.push(info);
                                }
                                else {
                                    throw new Error("Unsupported HubPool token: ".concat(JSON.stringify(info)));
                                }
                            }
                        };
                        this_1 = this;
                        for (_e = 0, tokenInfo_1 = tokenInfo; _e < tokenInfo_1.length; _e++) {
                            info = tokenInfo_1[_e];
                            _loop_1(info);
                        }
                        uniqueL1Tokens.forEach(function (token, i) {
                            _this.lpTokens[token] = { lastLpFeeUpdate: lpTokenInfo[i].lastLpFeeUpdate };
                        });
                        (_h = this.proposedRootBundles).push.apply(_h, events["ProposeRootBundle"]
                            .filter(function (event) { return !_this.configOverride.ignoredHubProposedBundles.includes(event.blockNumber); })
                            .map(function (event) {
                            return __assign(__assign({}, spreadEventWithBlockNumber(event)), { transactionHash: event.transactionHash });
                        }));
                        (_j = this.canceledRootBundles).push.apply(_j, events["RootBundleCanceled"].map(function (event) { return spreadEventWithBlockNumber(event); }));
                        (_k = this.disputedRootBundles).push.apply(_k, events["RootBundleDisputed"].map(function (event) { return spreadEventWithBlockNumber(event); }));
                        for (_f = 0, _g = events["RootBundleExecuted"]; _f < _g.length; _f++) {
                            event_3 = _g[_f];
                            if (this.configOverride.ignoredHubExecutedBundles.includes(event_3.blockNumber)) {
                                continue;
                            }
                            executedRootBundle = spreadEventWithBlockNumber(event_3);
                            l1Tokens = executedRootBundle.l1Tokens, runningBalances = executedRootBundle.runningBalances;
                            nTokens = l1Tokens.length;
                            // Safeguard
                            if (![nTokens, nTokens * 2].includes(runningBalances.length)) {
                                throw new Error("Invalid runningBalances length: ".concat(runningBalances.length, ". Expected ").concat(nTokens, " or ").concat(nTokens * 2, " for chain ").concat(this.chainId, " transaction ").concat(event_3.transactionHash));
                            }
                            executedRootBundle.runningBalances = runningBalances.slice(0, nTokens);
                            executedRootBundle.incentiveBalances =
                                runningBalances.length > nTokens ? runningBalances.slice(nTokens) : runningBalances.map(function () { return toBN(0); });
                            this.executedRootBundles.push(executedRootBundle);
                        }
                        // If the contract's current rootBundleProposal() value has an unclaimedPoolRebalanceLeafCount > 0, then
                        // it means that either the root bundle proposal is in the challenge period and can be disputed, or it has
                        // passed the challenge period and pool rebalance leaves can be executed. Once all leaves are executed, the
                        // unclaimed count will drop to 0 and at that point there is nothing more that we can do with this root bundle
                        // besides proposing another one.
                        if (pendingRootBundleProposal.unclaimedPoolRebalanceLeafCount > 0) {
                            mostRecentProposedRootBundle = this.proposedRootBundles[this.proposedRootBundles.length - 1];
                            this.pendingRootBundle = {
                                poolRebalanceRoot: pendingRootBundleProposal.poolRebalanceRoot,
                                relayerRefundRoot: pendingRootBundleProposal.relayerRefundRoot,
                                slowRelayRoot: pendingRootBundleProposal.slowRelayRoot,
                                proposer: pendingRootBundleProposal.proposer,
                                unclaimedPoolRebalanceLeafCount: pendingRootBundleProposal.unclaimedPoolRebalanceLeafCount,
                                challengePeriodEndTimestamp: pendingRootBundleProposal.challengePeriodEndTimestamp,
                                bundleEvaluationBlockNumbers: mostRecentProposedRootBundle.bundleEvaluationBlockNumbers.map(function (block) {
                                    // Ideally, the HubPool.sol contract should limit the size of the elements within the
                                    // bundleEvaluationBlockNumbers array. But because it doesn't, we wrap the cast of BN --> Number
                                    // in a try/catch statement and return some value that would always be disputable.
                                    // This catches the denial of service attack vector where a malicious proposer proposes with bundle block
                                    // evaluation block numbers larger than what BigNumber::toNumber() can handle.
                                    try {
                                        return block.toNumber();
                                    }
                                    catch (_a) {
                                        return 0;
                                    }
                                }),
                                proposalBlockNumber: mostRecentProposedRootBundle.blockNumber,
                            };
                        }
                        else {
                            this.pendingRootBundle = undefined;
                        }
                        this.currentTime = currentTime;
                        this.latestBlockNumber = latestBlockNumber;
                        this.firstBlockToSearch = update.searchEndBlock + 1; // Next iteration should start off from where this one ended.
                        this.isUpdated = true;
                        this.logger.debug({ at: "HubPoolClient::update", message: "HubPool client updated!", endBlock: latestBlockNumber });
                        return [2 /*return*/];
                }
            });
        });
    };
    // Returns end block for `chainId` in ProposedRootBundle.bundleBlockEvalNumbers. Looks up chainId
    // in chainId list, gets the index where its located, and returns the value of the index in
    // bundleBlockEvalNumbers. Returns 0 if `chainId` can't be found in `chainIdList` and if index doesn't
    // exist in bundleBlockEvalNumbers.
    HubPoolClient.prototype.getBundleEndBlockForChain = function (proposeRootBundleEvent, chainId, chainIdList) {
        var bundleEvaluationBlockNumbers = proposeRootBundleEvent.bundleEvaluationBlockNumbers;
        var chainIdIndex = chainIdList.indexOf(chainId);
        if (chainIdIndex === -1) {
            return 0;
        }
        // Sometimes, the root bundle event's chain ID list will update from bundle to bundle, so we need to check that
        // the bundle evaluation block number list is long enough to contain this index. We assume that chain ID's
        // are only added to the bundle block list, never deleted.
        if (chainIdIndex >= bundleEvaluationBlockNumbers.length) {
            return 0;
        }
        return bundleEvaluationBlockNumbers[chainIdIndex].toNumber();
    };
    HubPoolClient.prototype.updateFromJSON = function (hubPoolClientState) {
        var keysToUpdate = Object.keys(hubPoolClientState);
        this.logger.debug({
            at: "HubPoolClient",
            message: "Updating HubPool client from JSON",
            keys: keysToUpdate,
        });
        if (keysToUpdate.length === 0) {
            return;
        }
        if (!this.configStoreClient.isUpdated) {
            throw new Error("ConfigStoreClient not updated");
        }
        var _a = hubPoolClientState.l1TokensToDestinationTokens, l1TokensToDestinationTokens = _a === void 0 ? this.l1TokensToDestinationTokens : _a, _b = hubPoolClientState.l1Tokens, l1Tokens = _b === void 0 ? this.l1Tokens : _b, _c = hubPoolClientState.lpTokens, lpTokens = _c === void 0 ? this.lpTokens : _c, _d = hubPoolClientState.canceledRootBundles, canceledRootBundles = _d === void 0 ? this.canceledRootBundles : _d, _e = hubPoolClientState.disputedRootBundles, disputedRootBundles = _e === void 0 ? this.disputedRootBundles : _e, _f = hubPoolClientState.pendingRootBundle, pendingRootBundle = _f === void 0 ? this.pendingRootBundle : _f, _g = hubPoolClientState.crossChainContracts, crossChainContracts = _g === void 0 ? this.crossChainContracts : _g, _h = hubPoolClientState.l1TokensToDestinationTokensWithBlock, l1TokensToDestinationTokensWithBlock = _h === void 0 ? this.l1TokensToDestinationTokensWithBlock : _h, _j = hubPoolClientState.firstBlockToSearch, firstBlockToSearch = _j === void 0 ? this.firstBlockToSearch : _j, _k = hubPoolClientState.latestBlockNumber, latestBlockNumber = _k === void 0 ? this.latestBlockNumber : _k, _l = hubPoolClientState.currentTime, currentTime = _l === void 0 ? this.currentTime : _l, proposedRootBundles = hubPoolClientState.proposedRootBundles, executedRootBundles = hubPoolClientState.executedRootBundles;
        this.l1TokensToDestinationTokens = l1TokensToDestinationTokens;
        this.l1Tokens = l1Tokens;
        this.lpTokens = lpTokens;
        this.proposedRootBundles = proposedRootBundles
            ? proposedRootBundles.map(function (bundle) { return (__assign(__assign({}, bundle), { bundleEvaluationBlockNumbers: bundle.bundleEvaluationBlockNumbers.map(function (block) { return BigNumber.from(block); }) })); })
            : this.proposedRootBundles;
        this.canceledRootBundles = canceledRootBundles;
        this.disputedRootBundles = disputedRootBundles;
        this.executedRootBundles = executedRootBundles
            ? executedRootBundles.map(function (bundle) { return (__assign(__assign({}, bundle), { bundleLpFees: bundle.bundleLpFees.map(function (fee) { return BigNumber.from(fee); }), netSendAmounts: bundle.netSendAmounts.map(function (amount) { return BigNumber.from(amount); }), runningBalances: bundle.runningBalances.map(function (balance) { return BigNumber.from(balance); }), incentiveBalances: bundle.incentiveBalances.map(function (balance) { return BigNumber.from(balance); }) })); })
            : this.executedRootBundles;
        this.pendingRootBundle = pendingRootBundle;
        this.crossChainContracts = crossChainContracts;
        this.l1TokensToDestinationTokensWithBlock = l1TokensToDestinationTokensWithBlock;
        this.firstBlockToSearch = firstBlockToSearch;
        this.latestBlockNumber = latestBlockNumber;
        this.currentTime = currentTime;
        this.isUpdated = true;
    };
    HubPoolClient.prototype.toJSON = function () {
        return {
            deploymentBlock: this.deploymentBlock,
            chainId: this.chainId,
            eventSearchConfig: this.eventSearchConfig,
            configOverride: this.configOverride,
            firstBlockToSearch: this.firstBlockToSearch,
            latestBlockNumber: this.latestBlockNumber,
            currentTime: this.currentTime,
            l1TokensToDestinationTokens: this.l1TokensToDestinationTokens,
            l1Tokens: this.l1Tokens,
            lpTokens: this.lpTokens,
            proposedRootBundles: this.proposedRootBundles.map(function (bundle) {
                return JSON.parse(stringifyJSONWithNumericString(bundle));
            }),
            canceledRootBundles: this.canceledRootBundles,
            disputedRootBundles: this.disputedRootBundles,
            executedRootBundles: this.executedRootBundles.map(function (bundle) {
                return JSON.parse(stringifyJSONWithNumericString(bundle));
            }),
            pendingRootBundle: this.pendingRootBundle,
            crossChainContracts: this.crossChainContracts,
            l1TokensToDestinationTokensWithBlock: this.l1TokensToDestinationTokensWithBlock,
        };
    };
    return HubPoolClient;
}(BaseAbstractClient));
export { HubPoolClient };
//# sourceMappingURL=HubPoolClient.js.map