import { __assign, __awaiter, __generator, __spreadArray } from "tslib";
import EventEmitter from "events";
import { clientConfig } from "./config";
import { SpokePoolEventsQueryService } from "./services/SpokePoolEventsQueryService";
import { Logger } from "./adapters/logger";
import { BigNumber } from "ethers";
import { SpokePoolEventsQuerier } from "./adapters/web3";
import { TransfersRepository } from "./adapters/db/transfers-repository";
import { SpokePool__factory, } from "../typechain";
import { ethers } from "ethers";
export var TransfersHistoryEvent;
(function (TransfersHistoryEvent) {
    TransfersHistoryEvent["TransfersUpdated"] = "TransfersUpdated";
})(TransfersHistoryEvent || (TransfersHistoryEvent = {}));
var TransfersHistoryClient = /** @class */ (function () {
    function TransfersHistoryClient(config, logger, transfersRepository) {
        if (logger === void 0) { logger = new Logger(); }
        if (transfersRepository === void 0) { transfersRepository = new TransfersRepository(); }
        this.logger = logger;
        this.transfersRepository = transfersRepository;
        this.eventEmitter = new EventEmitter();
        this.web3Providers = {};
        this.spokePoolInstances = {};
        this.eventsQueriers = {};
        this.eventsServices = {};
        this.pollingIntervalSeconds = 15;
        this.pollingTimers = {};
        this.fetchingState = {};
        if (typeof config.pollingIntervalSeconds === "number") {
            this.pollingIntervalSeconds = config.pollingIntervalSeconds;
        }
        for (var _i = 0, _a = config.chains; _i < _a.length; _i++) {
            var chain = _a[_i];
            this.web3Providers[chain.chainId] = chain.provider;
            this.spokePoolInstances[chain.chainId] = SpokePool__factory.connect(chain.spokePoolContractAddr, this.web3Providers[chain.chainId]);
            this.eventsQueriers[chain.chainId] = new SpokePoolEventsQuerier(this.spokePoolInstances[chain.chainId], undefined, this.logger);
            clientConfig.spokePools[chain.chainId] = { lowerBoundBlockNumber: chain.lowerBoundBlockNumber || 0 };
        }
    }
    TransfersHistoryClient.prototype.setLogLevel = function (level) {
        this.logger.setLevel(level);
    };
    TransfersHistoryClient.prototype.startFetchingTransfers = function (depositorAddr) {
        return __awaiter(this, void 0, void 0, function () {
            var timer;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.fetchingState[depositorAddr] = "started";
                        this.initSpokePoolEventsQueryServices(depositorAddr);
                        return [4 /*yield*/, this.getEventsForDepositor(depositorAddr)];
                    case 1:
                        _a.sent();
                        // mark that we started fetching events for depositor address
                        // add polling if user didn't opted out for polling
                        if (this.pollingIntervalSeconds > 0) {
                            timer = this.pollingTimers[depositorAddr];
                            // prevent triggering multiple polling intervals for the same address
                            if (timer)
                                throw new Error("Address ".concat(depositorAddr, " is already monitored"));
                            timer = setInterval(function () {
                                _this.getEventsForDepositor(depositorAddr);
                            }, this.pollingIntervalSeconds * 1000);
                            this.pollingTimers[depositorAddr] = timer;
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    TransfersHistoryClient.prototype.stopFetchingTransfers = function (depositorAddr) {
        var timer = this.pollingTimers[depositorAddr];
        // mark that the fetching stopped for depositor address
        this.fetchingState[depositorAddr] = "stopped";
        if (timer) {
            clearInterval(timer);
            delete this.pollingTimers[depositorAddr];
        }
    };
    TransfersHistoryClient.prototype.on = function (event, cb) {
        this.eventEmitter.on(event, cb);
    };
    TransfersHistoryClient.prototype.initSpokePoolEventsQueryServices = function (depositorAddr) {
        var chainIds = Object.keys(this.spokePoolInstances).map(function (chainId) { return parseInt(chainId); });
        for (var _i = 0, chainIds_1 = chainIds; _i < chainIds_1.length; _i++) {
            var chainId = chainIds_1[_i];
            if (!this.eventsServices[depositorAddr]) {
                this.eventsServices[depositorAddr] = {};
            }
            if (!this.eventsServices[depositorAddr][chainId]) {
                this.eventsServices[depositorAddr][chainId] = new SpokePoolEventsQueryService(chainId, this.web3Providers[chainId], this.eventsQueriers[chainId], this.logger, depositorAddr !== "all" ? depositorAddr : undefined);
            }
        }
    };
    TransfersHistoryClient.prototype.getEventsForDepositor = function (depositorAddr) {
        return __awaiter(this, void 0, void 0, function () {
            var events, depositEvents, filledRelayEvents, speedUpDepositEvents, blockTimestampMap, filledTransfersCount, pendingTransfersCount, eventData;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, Promise.all(Object.values(this.eventsServices[depositorAddr]).map(function (eventService) { return eventService.getEvents(); }))];
                    case 1:
                        events = _a.sent();
                        depositEvents = events
                            .flat()
                            .reduce(function (acc, val) { return __spreadArray(__spreadArray([], acc, true), val.depositEvents, true); }, []);
                        filledRelayEvents = events
                            .flat()
                            .reduce(function (acc, val) { return __spreadArray(__spreadArray([], acc, true), val.filledRelayEvents, true); }, []);
                        speedUpDepositEvents = events.flat().reduce(function (acc, val) { return __spreadArray(__spreadArray([], acc, true), val.speedUpDepositEvents.map(function (event) { return (__assign(__assign({}, event), { 
                            // we have to enrich the event with this information in order to determine
                            // which exact `transfer` needs to be updated.
                            emittedFromChainId: val.emittedFromChainId })); }), true); }, []);
                        blockTimestampMap = events
                            .flat()
                            .reduce(function (acc, val) { return (__assign(__assign({}, acc), val.blockTimestampMap)); }, {});
                        depositEvents.map(function (e) { return _this.insertFundsDepositedEvent(e, blockTimestampMap[e.blockNumber]); });
                        filledRelayEvents.map(function (e) { return _this.insertFilledRelayEvent(e); });
                        speedUpDepositEvents.map(function (e) { return _this.handleSpeedUpDepositEvent(e, blockTimestampMap[e.blockNumber]); });
                        this.transfersRepository.aggregateTransfers();
                        filledTransfersCount = depositorAddr === "all"
                            ? this.transfersRepository.countAllFilledTransfers()
                            : this.transfersRepository.countFilledTransfers(depositorAddr);
                        pendingTransfersCount = depositorAddr === "all"
                            ? this.transfersRepository.countAllPendingTransfers()
                            : this.transfersRepository.countPendingTransfers(depositorAddr);
                        eventData = {
                            depositorAddr: depositorAddr,
                            filledTransfersCount: filledTransfersCount,
                            pendingTransfersCount: pendingTransfersCount,
                        };
                        // emit event only if the fetching wasn't stopped for depositor address.
                        // this is to prevent events from being triggered after the fetching was stopped
                        if (this.fetchingState[depositorAddr] === "started") {
                            this.eventEmitter.emit(TransfersHistoryEvent.TransfersUpdated, eventData);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    TransfersHistoryClient.prototype.insertFundsDepositedEvent = function (event, timestamp) {
        var args = event.args, transactionHash = event.transactionHash;
        var amount = args.amount, originToken = args.originToken, destinationChainId = args.destinationChainId, depositId = args.depositId, depositor = args.depositor, originChainId = args.originChainId, relayerFeePct = args.relayerFeePct;
        var transfer = {
            amount: BigNumber.from(amount),
            assetAddr: originToken,
            depositId: depositId,
            depositTime: timestamp,
            depositTxHash: transactionHash,
            destinationChainId: destinationChainId.toNumber(),
            filled: ethers.constants.Zero,
            sourceChainId: originChainId.toNumber(),
            status: "pending",
            fillTxs: [],
            initialRelayerFeePct: relayerFeePct,
            currentRelayerFeePct: relayerFeePct,
            speedUps: [],
        };
        this.transfersRepository.insertTransfer(originChainId.toNumber(), depositor, depositId, transfer);
    };
    TransfersHistoryClient.prototype.insertFilledRelayEvent = function (event) {
        var args = event.args, transactionHash = event.transactionHash;
        var totalFilledAmount = args.totalFilledAmount, depositor = args.depositor, depositId = args.depositId, originChainId = args.originChainId, relayerFeePct = args.relayerFeePct;
        this.transfersRepository.updateFilledAmount(originChainId.toNumber(), depositor, depositId, totalFilledAmount, transactionHash, relayerFeePct);
    };
    TransfersHistoryClient.prototype.handleSpeedUpDepositEvent = function (event, timestamp) {
        var args = event.args, transactionHash = event.transactionHash, emittedFromChainId = event.emittedFromChainId;
        var newRelayerFeePct = args.newRelayerFeePct, depositor = args.depositor, depositId = args.depositId;
        this.transfersRepository.updateRelayerFee(emittedFromChainId, depositor, depositId, newRelayerFeePct, transactionHash, timestamp);
    };
    TransfersHistoryClient.prototype.getFilledTransfers = function (depositorAddr, limit, offset) {
        if (depositorAddr === "all") {
            return this.transfersRepository.getAllFilledTransfers();
        }
        return this.transfersRepository.getFilledTransfers(depositorAddr, limit, offset);
    };
    TransfersHistoryClient.prototype.getPendingTransfers = function (depositorAddr, limit, offset) {
        if (depositorAddr === "all") {
            return this.transfersRepository.getAllPendingTransfers();
        }
        return this.transfersRepository.getPendingTransfers(depositorAddr, limit, offset);
    };
    return TransfersHistoryClient;
}());
export { TransfersHistoryClient };
//# sourceMappingURL=client.js.map